import pandas as pd
import logging

logger = logging.getLogger(__name__)

def log_header():
    header = """
888888888888         88  88888888ba   88888888ba   
     88              88  88      "8b  88      "8b  
     88              88  88      ,8P  88      ,8P  
     88  8b,dPPYba,  88  88aaaaaa8P'  88aaaaaa8P'  
     88  88P'   "Y8  88  88""""""'    88""""""'    
     88  88          88  88           88           
     88  88          88  88           88           
     88  88          88  88           88           


The Trajectory Iterative pKa Predictor (TrIPP) 
Written by: Christos Matsingos, Ka Fu Man, and Arianna Fornili 

If you are using TrIPP, please cite: 

-----------------------------------------------------------------
"""
    return header

def pka_statistics_table(df):
    """
    Generates a formatted string representation of pKa statistics from a DataFrame.
    Parameters
    ----------
    df: pd.DataFrame
        A DataFrame containing pKa values for different residues at different time points.
        Columns (labelled with residue identifiers) correspond to residues and 
        and rows to time points.
    Returns
    -------
    pka_statistics_table_in_str: str
        A string representation of the pKa statistics table.
    """
    pka_statistics = []
    for residue, pKaValues in df.items():
        if 'Time [ps]' == residue:
            continue
        pka_statistics.append([residue,
                            "{:.2f}".format(pKaValues.mean()),
                            "{:.2f}".format(pKaValues.median()),
                            "{:.2f}".format(pKaValues.std())])
    pka_statistics_table = pd.DataFrame(pka_statistics,
                                        columns=['Residue',
                                                'Mean',
                                                'Median',
                                                'Standard_Deviation'])
    tmp = pka_statistics_table.to_string(index=False).split('\n')
    tmp = [','.join(ele.split()) for ele in tmp]
    tmp = [element.replace('_',' ') for element in tmp]
    pka_statistics_table_in_str = '\r\n'.join(tmp)
    return pka_statistics_table_in_str

def trajectory_log(output_directory,
                   output_prefix, 
                   extract_buriedness_data,
                   mutation_selection, 
                   disulphide_cys_col,
                   optargs,
                   cores,
                   trajectory_slices,
                   start,
                   end):
    """
    Logs the parameters and results of the pKa calculation.
    Parameters
    ----------
    output_directory : str
        The directory where the output files are saved.
    output_prefix : str
        The prefix for the output files.
    extract_buriedness_data : bool
        Whether to extract buriedness (buried ratio) data. 
    mutation_selection : str
        Selection string for pseudo-mutations.
    disulphide_cys_col : list | None
        List of disulphide bonded cysteines.
    optargs : dict
        Optional arguments for the PROPKA calculation.
    cores : int
        The number of cores to use for the calculation.
    trajectory_slices : list
        A list of the trajectory slices that are processed in parallel.
    start : str
        Wall-clock start time.
    end : str
        Wall-clock end time.
    """
    if disulphide_cys_col:
        save_disulphide_pka = False
    else:
        save_disulphide_pka = True
        
    logger.info(f"""-----------------------------------------------------------------                

Start time: {start}
End time: {end}

PARAMETERS:
Output directory: {output_directory}
Output prefix: {output_prefix}
Number of cores: {cores}
Trajectory slices: {trajectory_slices}
Pseudo-mutations: {mutation_selection}
Extract buriedness: {extract_buriedness_data}
Save disulphide bond cysteines in csv: {save_disulphide_pka}
List of cysteines removed: {disulphide_cys_col}
PROPKA optional arguments: {optargs}

-----------------------------------------------------------------
""")
    df = pd.read_csv(f'{output_directory}/{output_prefix}_pka.csv')
    df.drop(columns=['Time [ps]'],inplace=True)
    chains = set([x.split(':')[-1] for x in df.columns])
    for chain in chains:
        df_chain = df[df.columns[df.columns.str.split(':').str[-1] == chain]]
        logger.info(f"""pKa Statistics for chain {chain}:
{pka_statistics_table(df_chain)}

-----------------------------------------------------------------
""")


"""
    @release_date  : $release_date
    @version       : $release_version
    @author        : Christos Matsingos, Ka Fu Man 
    
    This file is part of the TrIPP software
    (https://github.com/fornililab/TrIPP).
    Copyright (c) 2024 Christos Matsingos, Ka Fu Man and Arianna Fornili.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
from tqdm import tqdm

def greedy_clustering(cutoff, rmsd_matrix, frames, trajectory_names):
    """
    Function to run greedy clustering as described by Micheletti et al.
    (DOI: 10.1002/1097-0134(20000901)40:4<662::aid-prot90>3.0.co;2-f).
    
    Parameters
    ----------
    cutoff : float
        The distance cutoff for clustering.
    rmsd_matrix : np.ndarray
        The distance matrix (Euclidean distance) generated by the calculate_rmsd_matrix function.
    frames : list
        A list of frames corresponding to the points to be clustered.
    trajectory_names : np.ndarray
        An array of trajectory names corresponding to the clustering points.
        Generated by the create_clustering_matrix function.
    Returns
    -------
    labels: np.ndarray
        An array of cluster labels for each point in the feature matrix.
    cluster_centers: list
        Indices of the cluster centers within their individual trajectories (local indices). 
        Indices start from 0 for each trajectory.
    cluster_center_indices: list
        Indices of the cluster centers within the full feature matrix (global indices).
    cluster_centers_trajectories: list
        Trajectory names of the cluster centers.
    """

    # A copy of the RMSD matrix is made to not modify the original array
    current_matrix = rmsd_matrix.copy()
    n_structures = len(frames)

    # A labels array with the same number of elements as the total numebr of
    # structures is generated. Initially all structures are labelled to
    # belong to cluster 0.
    labels = np.zeros(n_structures, dtype=int)
    cluster = 1
    cluster_centers = []
    cluster_center_indices = []
    cluster_centers_trajectories = []

    for i in tqdm(range(n_structures)):
        # A boolean array neighbor_mask is generated to determine what RMSDs
        # are under the cutoff.
        neighbor_mask = current_matrix <= cutoff

        # The current_matrix array is checked to determine if all elements are
        # set to infinity. If the condition is true the clustering is complete.
        if np.all(np.isinf(current_matrix)):
            break

        else:
            # The amount of neighbors (structures within the cutoff)
            # is determined for each structure.
            neighbor_counts = np.sum(neighbor_mask, axis=0)
            # The structure with the most neighbors is determined to
            # be the center of the cluster
            central_structure = np.argmax(neighbor_counts)
            cluster_centers.append(frames[central_structure])
            cluster_centers_trajectories.append(trajectory_names[central_structure])
            cluster_center_indices.append(central_structure)
            # All structures neighboring the central structure are determined
            # and are assigned to the same cluster
            neighbors = np.ravel(neighbor_mask[central_structure, :])
            labels[neighbors] = cluster
            # The current_matrix (copy of the RMSD matrix) is altered to
            # remove elements belonging to the cluster from future clustering
            # by setting the RMSD values to inf.
            current_matrix[central_structure, :] = np.inf
            current_matrix[:, central_structure] = np.inf
            current_matrix[neighbor_mask[central_structure], :] = np.inf
            current_matrix[:, neighbor_mask[central_structure]] = np.inf
            cluster += 1  # The following cluster label is generated

    cluster_centers = list(np.ravel(np.array(cluster_centers)))
    cluster_center_indices = list(np.ravel(np.array(cluster_center_indices)))
    cluster_centers_trajectories = list(np.ravel(np.array(cluster_centers_trajectories)))
    # All labels are changed to fit the numbering of the other clustering 
    # methods.
    labels = (labels - 1)

    return labels, cluster_centers, cluster_center_indices, cluster_centers_trajectories



import pandas as pd 
import logging
import numpy as np

logger = logging.getLogger(__name__)

def output_df(output_directory, output_prefix, data, save_disulphide_pka, extract_buriedness_data):
    """
    Combines the results from all worker processes and saves the pKa and buriedness (buried ratio) values to CSV files.
    Parameters
    ----------
    output_directory : str
        The path to the directory where output files will be saved.
    output_prefix : str
        Output file prefix.
    data : list
        A list of data from all workers, where each worker's data is a dictionary
        generated by extract_pka_buriedness_data
    save_disulphide_pka : bool
        Whether to save pKa values for cysteines in disulphide bonds.
    extract_buriedness_data : bool
        Whether to extract and save buriedness data.
    Returns
    -------
    disulphide_cys_col : list or None
        A list of residue identifiers for disulphide-bonded cysteines if save_disulphide_pka is True, 
        otherwise None.
    """
    pka_full_data = []
    buriedness_full_data = []
    time_list = []
    for data_worker in data:
        for pka_buriedness_resname_worker_dict in data_worker:
            for time, pka_buriedness_resname_dict in pka_buriedness_resname_worker_dict.items():
                pka_full_data.append(pka_buriedness_resname_dict['pka_list'])
                buriedness_full_data.append(pka_buriedness_resname_dict['buriedness_list'])
                time_list.append(time)
    pka_df = pd.DataFrame(pka_full_data, columns=pka_buriedness_resname_dict['residue_identifier_list'])
    pka_df.insert(0,'Time [ps]',time_list)
    
    try:
        pka_df = pka_df.astype(float).sort_values('Time [ps]')
    except ValueError:
        contains_star = pka_df.applymap(lambda x: '*' in x if isinstance(x, str) else False)
        coupled_residues = ", ".join(list(pka_df.loc[:, contains_star.any()].columns))
        logging.info(f"""PROPKA detected that your system has the following coupled residues:
{coupled_residues}
run -d in the optargs to consider alternative state.
""")
        pka_df = pka_df.applymap(lambda x: x.replace('*','') if isinstance(x, str) else x)
        pka_df = pka_df.astype(float).sort_values('Time [ps]')
        
    if extract_buriedness_data:
        buriedness_df = pd.DataFrame(buriedness_full_data, columns=pka_buriedness_resname_dict['residue_identifier_list'])
        buriedness_df.insert(0,'Time [ps]',time_list)
        buriedness_df['Time [ps]'] = buriedness_df['Time [ps]'].astype(float)
        buriedness_df.sort_values('Time [ps]', inplace=True)

    if not save_disulphide_pka:
        disulphide_cys_col = pka_df.columns[(pka_df == 99.99).any()].to_list()
        pka_df.drop(disulphide_cys_col, axis=1, inplace=True)
        if extract_buriedness_data:
            buriedness_df.drop(disulphide_cys_col, axis=1, inplace=True)
    else:
        disulphide_cys_col = None
    
    pka_df.to_csv(f'{output_directory}/{output_prefix}_pka.csv', index=False)
    
    if extract_buriedness_data:
        buriedness_df.to_csv(f'{output_directory}/{output_prefix}_buriedness.csv', index=False)
        
    return disulphide_cys_col


"""
    @release_date  : $release_date
    @version       : $release_version
    @author        : Christos Matsingos, Ka Fu Man 
    
    This file is part of the TrIPP software
    (https://github.com/fornililab/TrIPP).
    Copyright (c) 2024 Christos Matsingos, Ka Fu Man and Arianna Fornili.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
"""


import MDAnalysis as mda
import numpy as np
import pandas as pd 
from tripp._pymol_template_ import gen_pymol_template 
from tripp._model_pka_values_ import model_pka_values 
from tripp._create_mda_universe_ import create_mda_universe 
import logging

class Visualization: 


    """ 
    Class to process and colour-map residue properties on a reference structure using PyMOL.
    
    Parameters:
    ----------
    topology_file: str
        The path of the topology file used for visualization
    pKa_file: str or list, default=None
        The path or a list of paths for the CSV file(s) containing the pKa values to be visualized.
    correlation_file: str, default=None
        The path of the CSV file generated by PCProjectionScreening() containing the correlation 
        coefficient between a given property (PC1 projection in the tutorial) and the pKa value 
        of each ionisable residue.
    start: int, default=0
        Start time (from pKa_file). The average pKa or pKa difference values will be calculated over the frames 
        with simulation times from start to end. Not applicable when correlation values are visualized.
    end: int, default=-1
        End time (from pKa_file). The average pKa or pKa difference values will be calculated over the frames 
        with simulation times from start to end. Not applicable when correlation values are visualized.
    """


    def __init__(self, 
                 topology_file,
                 pka_file=None,
                 correlation_file=None,
                 start = 0, 
                 end = -1):
        self.logger = logging.getLogger()
        if self.logger.hasHandlers:
            for handler in self.logger.handlers[:]:
                self.logger.removeHandler(handler)
        logging.Formatter('%(message)s')
        self.logger.setLevel(logging.WARNING)
        
        self.topology_file = topology_file
        if correlation_file != None:
            self.correlation_file = pd.read_csv(correlation_file)
        self.u = create_mda_universe(topology_file=self.topology_file, trajectory_file=None) 
        
        if pka_file:
            if type(pka_file) == list: 
                pka_file_l = [] 
                for file in pka_file: 
                    pka_file_l.append(pd.read_csv(file)) 
                pka_values = pd.concat(pka_file_l) 
            
            else: 
                pka_values = pd.read_csv(pka_file)
            # If user wants to use part of the trajectory to calculate mean/difference_to_model_value.
            self.start = start
            if end == -1:
                self.end = pka_values['Time [ps]'].iloc[end]
            else:
                self.end = end
            self.pka_values = pka_values[pka_values['Time [ps]'].isin(np.arange(self.start,self.end+1))]
            
            # Reorder the values_df so that C- and N+ go to the end of the dataframe.
            N_C_cols = [col for col in self.pka_values.columns if 'C-' in col or 'N+' in col]
            Rest_cols = [col for col in self.pka_values.columns if col not in N_C_cols]
            self.pka_values = self.pka_values[Rest_cols + N_C_cols]
            
    def gen_pse(self, 
                pymol_path, 
                output_directory,
                output_prefix,
                coloring_method='mean', 
                lower_limit=0, 
                upper_limit=14,
                correlation_threshold=None,
                color_palette='red_white_blue'): 
        """
        Generates a PyMOL session file.
        
        Parameters
        ----------
        pymol_path: str
            Path to PyMOL. The script will spawn a subprocess shell to run a 
            python script in PyMOL.
        output_directory : str
            The directory where the output files are saved.
        output_prefix: str
            Prefix of the output PyMOL file (.pse). The prefix will be combined 
            with the coloring_method ('mean', 'difference_to_model_value' or 'correlation') 
            to give pse_output_filename.
        coloring_method: str, int, or float, default 'mean'
            Determines the property visualised as colour gradient. 
            If set to 'mean', the average pKa value calculated over all frames is used. 
            If set to 'difference_to_model_value', the difference between the PROPKA predicted 
            and model pKa value is used. 
            If a specific simulation time is indicated, the pKa values predicted for the 
            corresponding frames are visualised. 
            If set to 'correlation', the correlation coefficient between a time-dependent property
            and the pKa time evolution will be visualised. In this case a correlation_file will need
            to be provided and specified in the Visualization class.
        lower_limit: int or float, default=0
            Lower bound of the color scale in the PyMOL session. Residues with values
            below this threshold are assigned the gradient’s minimum color.
        upper_limit: int or float, default=14
            Upper bound of the color scale in the PyMOL session. Residues with values
            above this threshold are assigned the gradient’s maximum color.
        correlation_threshold: float, default=None
            If the coloring_method is set to 'correlation', this parameter can be used to
            filter the correlation values. Only residues with an absolute correlation value
            above the threshold will be included in the PyMOL session.
        color_palette: str, default='red_white_blue'
            Colour palette. The default is set to 'red_white_blue'. See PyMOL spectrum
            for allowed colour palettes (a three-colour palette is recommended).
        """

        # Calculation of values depending on colouring method
        if coloring_method == 'mean':
            del self.pka_values['Time [ps]']
            values_df = self.pka_values.mean(axis=0).to_frame(name='pKa').reset_index(names='Residue')
            tempfactors_output_topology_file = f"{output_directory}/{output_prefix}_mean.pdb"

        elif coloring_method == 'difference_to_model_value':
            del self.pka_values['Time [ps]']
            pka_values_mean = self.pka_values.mean(axis=0)
            for residue in pka_values_mean.keys():
                if 'N+' in residue:
                    pka_values_mean[residue] = pka_values_mean[residue] - model_pka_values['NTR']
                elif 'C-' in residue:
                    pka_values_mean[residue] = pka_values_mean[residue] - model_pka_values['CTR']
                else:
                    pka_values_mean[residue] = pka_values_mean[residue] - model_pka_values[residue[0:3]]
            values_df = pka_values_mean.to_frame(name='pKa').reset_index(names='Residue')
            tempfactors_output_topology_file = f"{output_directory}/{output_prefix}_difference_to_model_value.pdb"


        elif coloring_method == 'correlation':
            correlation_df = self.correlation_file
            if correlation_threshold is not None:
                correlation_df = correlation_df[abs(correlation_df['Correlation']) > correlation_threshold]
            values_df = correlation_df[['Residue', 'Correlation']]
            tempfactors_output_topology_file = f"{output_directory}/{output_prefix}_correlation.pdb"

        elif type(coloring_method) == int or type(coloring_method) == float:

            if type(self.pka_file) == list and len(self.pka_file) != 1:
                raise Exception('Please select only one CSV file when selecting a specific time step.')

            else:
                self.pka_values = self.pka_values[self.pka_values['Time [ps]'] == coloring_method]
                del self.pka_values['Time [ps]']
                values_df = self.pka_values.mean(axis=0).to_frame(name='pKa').reset_index(names='Residue')
                tempfactors_output_topology_file = f"{output_directory}/{output_prefix}_time{coloring_method}.pdb"
        else:
            raise SyntaxError(f"{coloring_method} is not allowed for coloring_method, allowed values are 'mean', 'difference_to_model_value', or 'correlation'.")
        
        # Unpacking the residues and values (pka or correlation) from values_df.
        # Looping through them to assign the value onto the tempfactor of ionisable
        # residues. The topology_file with the tempfactor is written as pdb and a PyMOL
        # session is generated as .pse.
        residue_identifiers, values = (columns for _, columns in values_df.items())
        for residue_identifier, value in zip(residue_identifiers, values):
            residue = residue_identifier.split(':')[0]
            chain = residue_identifier.split(':')[-1]
            if 'N+' in residue or 'C-' in residue:
                resid = int(residue[2:])
            else:
                resid = int(residue[3:])

            rounded_predicted_pka = round(value, 2)
            if 'N+' in residue:
                # Only tempfactor of NtermCap_atom_name of the first residue will be written
                ag = self.u.select_atoms(f'((around 2 (chainID {chain} and resid {resid} and backbone and type N)) and chainID {chain} and resid {resid}) and type H or (chainID {chain} and resid {resid} and backbone and type N)')
            elif 'C-' in residue:
                # Only tempfactor of CtermCap_atom_name of the last residue will be written
                ag = self.u.select_atoms(f'((around 2 (chainID {chain} and resid {resid} and backbone and type C and not name CA)) and chainID {chain} and resid {resid}) and type O or (chainID {chain} and resid {resid} and backbone and type C and not name CA)')
            else:
                # All atoms in the ionisable residue will be written
                ag = self.u.select_atoms(f'(chainID {chain} and resid {resid} and not backbone) or (chainID {chain} and resid {resid} and name CA)')
            ag.tempfactors = np.full(ag.tempfactors.shape,rounded_predicted_pka)
        ag = self.u.select_atoms('all')
        ag.write(tempfactors_output_topology_file)
        pse_output_filename = tempfactors_output_topology_file.replace('.pdb', '.pse')
        gen_pymol_template(tempfactors_output_topology_file,
                           pymol_path, 
                           pse_output_filename, 
                           values_df, 
                           lower_limit, 
                           upper_limit, 
                           color_palette)


"""
    @release_date  : $release_date
    @version       : $release_version
    @author        : Christos Matsingos, Ka Fu Man 
    
    This file is part of the TrIPP software
    (https://github.com/fornililab/TrIPP).
    Copyright (c) 2024 Christos Matsingos, Ka Fu Man and Arianna Fornili.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
import logging
import numpy as np
from tripp._correction_dictionary_ import corrected_amino_acids

logger = logging.getLogger(__name__)
    
def check_resname_HETATM(non_protein_ag):
    """
    Checks if the resnames in the non-protein atoms universe are compatible with PROPKA.
    Parameters
    ----------
    non_protein_ag : MDAnalysis.AtomGroup
        The AtomGroup (MDAnalysis selection syntax) for residues with resnames not found in either the built-in or user-defined conversion dictionaries.
        
    Raises
    ------
    NameError
        If there are resnames that are not recognized by PROPKA, and their record type
        is not 'HETATM', an exception will be raised.
    """
    compatible_resnames = np.unique(list(corrected_amino_acids.values()))
    incorrect_resnames=[]
    for resname, record_types in zip(non_protein_ag.residues.resnames,
                                     non_protein_ag.residues.record_types):
        if resname not in compatible_resnames and np.all(record_types != 'HETATM'):
            incorrect_resnames.append(f'{resname}')
    if len(incorrect_resnames) > 0:
        raise NameError(f"""Your system still contains resname(s) not recognised by PROPKA: {', '.join(incorrect_resnames)}
For amino acids, please use the custom_resname_correction argument to indicate valid PROPKA resnames for the residues indicated above.
For ligands, please use the hetatm_resname argument to convert the record type of the ligand to HETATM.""")
    else:
        logger.info('Resname and record type check passed.\n')
        
def check_terminal_oxygens(universe):
    """
    Checks if the C-terminal oxygen atoms in the universe are named 'O' and 'OXT
    Parameters
    ----------
    universe : MDAnalysis.Universe
        The MDAnalysis universe containing the system to be checked.
    Raises
    ------
    NameError
        If no C-terminal oxygen atoms are found with the names 'O' and 'OXT',
        an exception will be raised.
    """
    terminals = []
    for index, name in zip(universe.residues.resindices, universe.residues.names):
        if np.isin('O', name).any() and np.isin('OXT', name).any():
            ag = universe.select_atoms(f'resindex {index}')
            terminals.append(f'{ag.residues.resnames[0]}{ag.residues.resids[0]}')
    if len(terminals) > 0:
        logger.info(f"""Terminal oxygen check passed, involving: 
{', '.join(terminals)}\n""")
    else:
        raise NameError('No terminal oxygen atom named O and OXT was found, please either modify your topology_file or use the custom_terminal_oxygens argument')


"""
    @release_date  : $release_date
    @version       : $release_version
    @author        : Christos Matsingos, Ka Fu Man 
    
    This file is part of the TrIPP software
    (https://github.com/fornililab/TrIPP).
    Copyright (c) 2024 Christos Matsingos, Ka Fu Man and Arianna Fornili.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

from sklearn.decomposition import PCA

def pca(clustering_matrix): 
    """
    Performs PCA on the feature matrix to reduce dimensionality. 
    Parameters
    ----------
    clustering_matrix: np.ndarray
        The clustering (feature) matrix to be transformed.
    Returns
    -------
    n_components: int
        The number of principal components (PCs) used for dimensionality reduction. 
        It is determined as the minimum number of principal components that explain
        at least 90% of the total variance. 
    cummulative_variance: float
        The cumulative variance explained by the selected principal components, expressed as a percentage.
    clustering_matrix_transformed: np.ndarray
        The transformed clustering matrix after PCA.
    """
    pca_class = PCA() 
    pca_class.fit(clustering_matrix) 
    var_ratio = pca_class.explained_variance_ratio_ 
    n_components = 0
    cummulative_variance = 0
    for item in var_ratio: 
        n_components += 1
        cummulative_variance += item 
        if cummulative_variance >= 0.9: 
            break 
    
    cummulative_variance = round(cummulative_variance*100, 2) 
    
    pca_class = PCA(n_components=n_components) 
    clustering_matrix_transformed = pca_class.fit_transform(clustering_matrix) 

    return n_components, cummulative_variance, clustering_matrix_transformed 


import MDAnalysis as mda 
from tripp._edit_pdb_ import mutate 
from propka import run 
from tripp._extract_pka_file_data_ import extract_pka_buriedness_data 
import os
import glob
import logging
import io

def pka_iterator(trajectory_slice, universe,
                 output_directory, mutation_selections,
                 optargs=[]):
    """
    Function to run propka.run.single on the distributed trajectory slice.
    
    Parameters
    ----------
    trajectory_slices: list of int
        Trajectory slices from the Trajectory class initialisation.
    universe: MDAnalysis.universe object
        Modified MDAnalysis universe from the Trajectory class initialisation.
    output_directory: str
        Directory to write the PROPKA output files to.
    mutation_selections: str
        Selection string in MDAnalysis format (only for pseudo-mutations) 
    optargs: list of str, default=[]
        PROPKA predictions can be run with optional arguments
        (see https://propka.readthedocs.io/en/latest/command.html).
        For example, if optargs is set to `["-k"]`, propka will run with the -k flag
        (protons from the input file are kept).
    """
    # Redirect warning from propka.group to a file
    logger = logging.getLogger('propka')
    logger.propagate = False 
    logger.setLevel(logging.WARNING)
    log_capture_string = io.StringIO()
    handler = logging.StreamHandler(log_capture_string)
    logger.addHandler(handler)
    log_contents = None
    
    pid = os.getpid()
    
    temp_name = f'{output_directory}/.temp_{pid}'

    start = trajectory_slice[0]
    end = trajectory_slice[1]
    
    cwd = os.getcwd()
    data = []
    for ts in universe.trajectory[start:end]:
        if mutation_selections is not None:
            mutate(universe, mutation_selections, temp_name)
        else:
            with mda.Writer(f'{temp_name}.pdb') as w:
                w.write(universe)
        os.chdir(output_directory)
        run.single(f'.temp_{pid}.pdb', optargs=optargs)
        
        if log_capture_string.getvalue():
            log_contents = (f"PROPKA warning raised for frame {ts.frame}:\n" + 
                            log_capture_string.getvalue()+
                            '\n')
        os.chdir(cwd)

        time = ts.time
        # Extract pKa and buriedness data from the generated .pka file
        # and append it to the data list.
        temp_pka_file = glob.glob(f'{temp_name}*.pka')[0] # .pka could have different suffixes ie: _alt_state.pka
        data_dictionary = extract_pka_buriedness_data(temp_pka_file, time=time)
        data.append(data_dictionary)

        os.remove(f'{temp_name}.pdb') 
        os.remove(temp_pka_file)
    
    logger.removeHandler(handler)
    log_capture_string.close()
    
    return data, log_contents



import numpy as np 

def extract_pka_buriedness_data(file, time):
    """
    Extracts pKa and buriedness (buried ratio) values from a PROPKA output file.
    Parameters
    ----------
    file: str
        The path to the PROPKA temporary output file.
    time: str
        The time of the frame processed by PROPKA, used as a key in the returned dictionary.
    Returns
    -------
    data: dict
        A nested dictionary where each time point is mapped to a dictionary containing the following keys:
        - 'residue_identifier_list': an array of residue identifiers in the format 'RESID:CHAIN'.
        - 'pka_list': an array of pKa values corresponding to the residues.
        - 'buriedness_list': an array of buriedness values corresponding to the residues.
        Elements in the three arrays are in the same order (element i from each array refers to the same residue)
    """
    compatible_resnames = ['ALA', 'ARG', 'ASN', 'ASP', 'CYS', 'GLN', 'GLU', 
                           'GLY', 'HIS', 'ILE', 'LEU', 'LYS', 'MET', 'PHE', 
                           'PRO', 'SER', 'THR', 'TRP', 'TYR', 'VAL', 'N+', 'C-'] # N+ and C- added for N- and C-termini in the PROPKA file
    pkafile = open(file, 'r')
    residue_identifier_list = []
    pka_list = [] 
    buriedness_list = []
    for line in pkafile: 
        line_processed = line.rstrip() 
        line_list = line_processed.strip().split()
        # identify the line containing the pka, only considering lines with amino acids (compatible_resnames).
        if len(line_list) > 15 and line_list[0] in compatible_resnames:
            residue_identifier_list.append(line_list[0]+line_list[1]+':'+line_list[2])
            pka_list.append(line_list[3])
            buriedness_list.append(line_list[4])

    data = {time: {'residue_identifier_list':np.array(residue_identifier_list),
                  'pka_list':np.array(pka_list),
                  'buriedness_list':np.array(buriedness_list)}}
    return data


import MDAnalysis as mda 
import numpy as np 
from tripp._correction_dictionary_ import corrected_amino_acids, corrected_atom_names
from tripp._propka_input_check_ import check_resname_HETATM, check_terminal_oxygens
import logging
import pandas as pd

logger = logging.getLogger(__name__)
def create_mda_universe(topology_file, trajectory_file):
    """
    Creates an MDAnalysis Universe object from a topology file and an 
    optional trajectory file. Checks for chain identity and formal 
    charges, and adds default values if necessary. Note that the 
    check for chain identity is not exhaustive.
    
    Parameters
    ----------
    topology_file : str
        Path to the topology file.
    trajectory_file : str or None
        Path to the trajectory file. If None, only the topology is loaded.
    Return
    -------
    universe : MDAnalysis.universe
        MDAnalysis Universe object containing the topology and trajectory 
        data.
    Raises
    ------
    Exception
        If the topology file contains duplicated residue IDs with empty chain ID 
        or the same chain ID, an exception is raised
    """
    if trajectory_file is None:
        universe = mda.Universe(topology_file)
    
    else: 
        universe = mda.Universe(topology_file, trajectory_file) 
    
    topology = universe._topology
    _, counts = np.unique(universe.residues.resids, return_counts=True)
    # Check if chainID of any residue is empty or not, 
    # if so default chain A for the whole system.
    if '' in topology.chainIDs.values and (counts == 1).all():
        topology.chainIDs.values = np.full(len(topology.chainIDs.values),
                                           'A',
                                           dtype=str) 
        logger.info(f"""Your topology file contains no chain identity for at least one atom. 
Will add chain A for your whole system by default""")
    elif '' in topology.chainIDs.values and (counts > 1).any():
        raise Exception('Your system contains duplicated residue IDs and at least one atom does not have a chain ID, please assign chain IDs before running TrIPP.')
    
    elif len(set(topology.chainIDs.values)) == 1 and (counts > 1).any():
        raise Exception('Your system contains duplicated resid, but all atoms in your system have the same chain ID, please re-assign the chain IDs before running TrIPP')
        
    if hasattr(universe.atoms, 'formalcharges') is False:
        universe.add_TopologyAttr('formalcharges', np.full(len(topology.chainIDs.values), 0, dtype=float))
        logger.info("""Your topology file contains no formal charges. 
Will set formal charges to 0 for your whole system.""")
    
    added_universe = universe.copy()
    return added_universe

def create_propka_compatible_universe(universe,
                                      hetatm_resname,
                                      custom_terminal_oxygens,
                                      custom_resname_correction):
    """
    Converts the universe from create_mda_universe function 
    into a PROPKA compatible format, including the resname of standard amino 
    acids and C-terminal oxygen atom names. The record type of a residue 
    can also be modified by the user (useful for ligands). Checks 
    are performed on resname, record type and C-terminal oxygen atom names. See 
    _propka_input_check_ for more information.
    
    Parameters
    ----------
    universe: MDAnalysis.universe
        MDAnalysis universe generated by create_mda_universe.
    hetatm_resname: str, list, default=None
        PDB residue name(s) for non-protein molecules that we want to 
        be taken into account in the pKa calculation. Their record type
        will be set to 'HETATM'. See log for info if an error is raised.
    custom_terminal_oxygens: list, default=None
        PROPKA only recognizes C-terminal oxygen atoms named O and OXT. 
        If different, either provide a list of length 2 containing the 
        names of your C-terminal oxygen atoms, or ser to None for correction 
        according to our pre-defined dictionary 
        (see _correction_dictionary_.py for the pre-defined names).
    custom_resname_correction: dict, default=None
        dictionary of custom protein residue names not included in
        the hard-coded TrIPP dictionary (tripp._correction_dictionary_.py). 
        Can be given as e.g. {'XXX':'ASP'}, where 'XXX' is the residue
        name in the PDB file and 'ASP' is the corresponding PROPKA name.
    Returns
    -------
    corrected_universe: MDAnalysis.universe
        MDAnalysis universe that is compatible with PROPKA, including
        20 standard amino acid resnames, O and OXT for C-terminal oxygen atoms, 
        HETATM for non-protein atoms (when hetatm_resname is used).
    """
    corrected_universe = universe.copy()
    # Make a local copy of corrected_amino_acids for temporary changes
    local_corrected_amino_acids = corrected_amino_acids.copy()
    if custom_resname_correction is not None:
        for current_resname, correct_resname in custom_resname_correction.items():
            local_corrected_amino_acids[current_resname] = correct_resname        
    # Correct resname to propka compatible with predefined list.
    corrected_resnames = []
    corrected_resnames_trace = []
    for resname, residx in zip(corrected_universe.residues.resnames,
                              corrected_universe.residues.resindices):
        if resname in local_corrected_amino_acids:
            corrected_resnames_trace.append(f'{resname}->{local_corrected_amino_acids[resname]}')
            resname = local_corrected_amino_acids[resname]
        # Special care taken for MSE -> MET, where the SE atomname needs to be modified as well.
        elif resname == 'MSE':
            corrected_resnames_trace.append(f'{resname}->MET (SE->SD)')
            resname = 'MET'
            atom_names = corrected_universe.select_atoms(f'resindex {residx}').atoms.names
            atom_types = corrected_universe.select_atoms(f'resindex {residx}').atoms.types
            corrected_universe.select_atoms(f'resindex {residx}').atoms.names = np.char.replace(atom_names.astype(str), 'SE', 'SD')
            corrected_universe.select_atoms(f'resindex {residx}').atoms.types = np.char.replace(atom_types.astype(str), 'SE', 'S')

        corrected_resnames.append(resname)
    
    corrected_universe.residues.resnames = corrected_resnames
    
    if len(corrected_resnames_trace) > 0:
        corrected_resnames_trace = '\n'.join(set(corrected_resnames_trace))
        logger.info(f"""The resnames of the following residues have been modified to be compatible with PROPKA:
{corrected_resnames_trace}""")
    
    # When hetatm_resname is provided, for example if the system contains ligand, 
    # the record type will be corrected to HETATM.
    if hetatm_resname is not None:
        if isinstance(hetatm_resname, list):
            hetatm_resname = ' '.join([str(i) for i in hetatm_resname])
        correction_ag = corrected_universe.select_atoms(f'resname {hetatm_resname}')
        correction_ag.atoms.record_types = np.full(len(correction_ag.atoms.record_types),'HETATM')
        corrected_hetatm = []
        for resid, resname in zip(correction_ag.residues.resids, correction_ag.residues.resnames):
            corrected_hetatm.append(f'{resname}{resid}')
        logger.info(f"""The record type of the following residues has been modified to HETATM:
{', '.join(corrected_hetatm)}""")
    
    propka_compatible_amino_acids = ' '.join(np.unique(list(local_corrected_amino_acids.values())))
    check_resname_HETATM(corrected_universe.select_atoms(f"not resname {propka_compatible_amino_acids} {hetatm_resname}"))
    
    # When custom_terminal_oxygens is not provided by user as list of string,
    # the terminal oxygens will be renamed according to the dictionary
    # in _correction_dictionary_.corrected_atom_names.
    if custom_terminal_oxygens is None:
        corrected_universe, corrected_terminal_oxygens_trace = modifiy_terminal_oxygens(corrected_universe, corrected_atom_names)
        logger.info(f"""C-Terminal oxygen atom names will be modified according to our predefined dictionary:
{corrected_terminal_oxygens_trace}""")
        
    # Check if argument is a list
    elif not isinstance(custom_terminal_oxygens,list):
        raise TypeError(f'custom_terminal_oxygens argument must be a list, not {type(custom_terminal_oxygens)}')
    
    # Check if argument is a list of string
    elif not all(isinstance(element,str) for element in custom_terminal_oxygens):
        raise TypeError(f'custom_terminal_oxygens argument must be a list of strings, at least one of your elements is not a string')
    
    # Check if argument is a list of length 2
    elif len(custom_terminal_oxygens) != 2:
        raise ValueError(f'Length of custom_terminal_oxygens should be 2, but got {len(custom_terminal_oxygens)}')

    # When custom_terminal_oxygens is provided by the user as a list of strings,
    # the C-terminal oxygen atoms will be renamed as O and OXT.
    elif isinstance(custom_terminal_oxygens, list) and len(custom_terminal_oxygens) == 2:
        correct_terminal_oxygens = ['O', 'OXT']
        to_be_corrected_dict = {i:j for i,j in zip(custom_terminal_oxygens, correct_terminal_oxygens)}
        corrected_universe, corrected_terminal_oxygens_trace = modifiy_terminal_oxygens(corrected_universe, to_be_corrected_dict)
        
        logger.info(f"""custom_terminal_oxygens supplied and modified:
{corrected_terminal_oxygens_trace}""")
    
    check_terminal_oxygens(corrected_universe)
        
    return corrected_universe

def modifiy_terminal_oxygens(corrected_universe, to_be_corrected_dict):
    """
    Function to modify the C-terminal oxygen atom names in the corrected_universe 
    based on to_be_corrected_dict.
    
    Parameters
    ----------
    corrected_universe: MDAnalysis.universe
        Universe to be modified.
    to_be_corrected_dict: dict
        Dictionary where the C-terminal oxygen atom names in the input topology 
        are the keys, while "O" or "OXT" are used as values.
        Provided through corrected_atom_names or by the user (via custom_terminal_oxygens).
    Return
    ----------
    corrected_universe: MDAnalysis.universe
        Universe with modified C-terminal oxygen atom names.
    corrected_terminal_oxygens_trace: str
        Information on the processed C-terminal oxygen atoms.
    """
    ag = corrected_universe.select_atoms('protein')
    corrected_terminal_oxygens_trace = []
    for index, name in zip(ag.residues.resindices, ag.residues.names):
        if np.isin([i.strip() for i in list(to_be_corrected_dict.keys())], name).any():
            corrected_name = pd.Series(name).replace(to_be_corrected_dict).to_numpy()
            ag_to_correct = corrected_universe.select_atoms(f'resindex {index}')
            before_correction = ag_to_correct.atoms.names.copy()
            ag_to_correct.atoms.names = corrected_name
            after_correction = ag_to_correct.atoms.names.copy()
            pnames = np.setdiff1d(before_correction, after_correction)
            anames = np.setdiff1d(after_correction, before_correction)
            resid = ag_to_correct.residues.resids[0]
            resname = ag_to_correct.residues.resnames[0]
            corrected_terminal_oxygens_trace.append(f"{resname}{resid}: {', '.join([f'{pname} -> {aname.strip()}' for pname,aname in zip(pnames,anames)])}")
    corrected_terminal_oxygens_trace = '\n'.join(corrected_terminal_oxygens_trace)
    return corrected_universe, corrected_terminal_oxygens_trace


# Contains residue names recognised by MDAnalysis and
# associates them with residue names recognised by PROPKA
corrected_amino_acids = {
    "ARGN": "ARG",
    "ASN1": "ASN",
    "ASPH": "ASP",
    "ASPP": "ASP",
    "CALA": "ALA",
    "CARG": "ARG",
    "CASF": "ASP",
    "CASN": "ASN",
    "CASP": "ASP",
    "CCYS": "CYS",
    "CCYX": "CYS",
    "CGLN": "GLN",
    "CGLU": "GLU",
    "CGLY": "GLY",
    "CHID": "HIS",
    "CHIE": "HIS",
    "CHIP": "HIS",
    "CILE": "ILE",
    "CLEU": "LEU",
    "CLYS": "LYS",
    "CMET": "MET",
    "CPHE": "PHE",
    "CPRO": "PRO",
    "CSER": "SER",
    "CTHR": "THR",
    "CTRP": "TRP",
    "CTYR": "TYR",
    "CVAL": "VAL",
    "CYS1": "CYS",
    "CYS2": "CYS",
    "CYSH": "CYS",
    "GLUH": "GLU",
    "GLUP": "GLU",
    "HIS1": "HIS",
    "HIS2": "HIS",
    "HISA": "HIS",
    "HISB": "HIS",
    "HISD": "HIS",
    "HISE": "HIS",
    "HISH": "HIS",
    "LYSH": "LYS",
    "NALA": "ALA",
    "NARG": "ARG",
    "NASN": "ASN",
    "NASP": "ASP",
    "NCYS": "CYS",
    "NCYX": "CYS",
    "NGLN": "GLY",
    "NGLU": "GLU",
    "NGLY": "GLY",
    "NHID": "HIS",
    "NHIE": "HIS",
    "NHIP": "HIS",
    "NILE": "ILE",
    "NLEU": "LEU",
    "NLYS": "LYS",
    "NMET": "MET",
    "NPHE": "PHE",
    "NPRO": "PRO",
    "NSER": "SER",
    "NTHR": "THR",
    "NTRP": "TRP",
    "NTYR": "TYR",
    "NVAL": "VAL",
    "ASF": "ASP",
    "ASH": "ASP",
    "CYM": "CYS",
    "CYN": "CYS",
    "CYX": "CYS",
    "GLH": "GLU",
    "HID": "HIS",
    "HIE": "HIS",
    "HIP": "HIS",
    "HSD": "HIS",
    "HSE": "HIS",
    "HSP": "HIS",
}
# Dictionary for correcting C-terminal oxygen atoms to O and OXT
# for GROMOS, AMBER, and CHARMM force fields.
corrected_atom_names = {
    "O1": "O",  # COO- (GROMOS)
    "O2": "OXT",  # COO- (GROMOS)
    "OT": "OXT",  # COOH (GROMOS)
    "OC1": "O",  # COO- (AMBER)
    "OC2": "OXT",  # COO- (AMBER)
    "OT1": "O",  # COO- (CHARMM)
    "OT2": "OXT",  # COO- (CHARMM)
}



"""
    @release_date  : $release_date
    @version       : $release_version
    @author        : Christos Matsingos, Ka Fu Man

    This file is part of the TrIPP software
    (https://github.com/fornililab/TrIPP).
    Copyright (c) 2024 Christos Matsingos, Ka Fu Man and Arianna Fornili.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

from sklearn_extra.cluster import KMedoids
import numpy as np


def kmedoids_clustering(
    n_clusters,
    metric,
    method,
    init,
    max_iter,
    random_state,
    clustering_matrix,
    frames,
    trajectory_names,
):
    """
    Function to run K-Medoids clustering from sklearn_extra.
    Standard K-Medoids parameters can be found in the sklearn_extra documentation:
    https://scikit-learn-extra.readthedocs.io/en/stable/generated/sklearn_extra.cluster.KMedoids.html
    
    Parameters
    ----------
    n_clusters: int
        The number of output clusters.
    metric: str
        The metric to use for distance computation.
    method: str
        The algorithm used for clustering.
    init: str
        The medoid initialization method.
    max_iter: int
        The maximum number of iterations for the clustering algorithm.
    random_state: int
        Random seed used in the medoid initialisation if init = 'random'.
    clustering_matrix: np.ndarray
        The clustering matrix (feature matrix) created by the create_clustering_matrix function.
    frames: np.ndarray
        A list of frames corresponding to the points to be clustered.
    trajectory_names: np.ndarray
        An array of names of the trajectories to be clustered.
    Returns
    -------
    labels: np.ndarray
        An array of cluster labels for each point in the feature matrix.
    cluster_centers: list
        Indices of the cluster centers (medoids) within their individual trajectories (local indices). 
        Indices start from 0 for each trajectory.
    cluster_center_indices: list
        Indices of the cluster centers (medoids) within the full feature matrix (global indices).
    cluster_centers_trajectories: list
        Trajectory names of the cluster centers (medoids).
    """

    kmedoids_clustering = KMedoids(
        n_clusters=n_clusters,
        metric=metric,
        method=method,
        init=init,
        max_iter=max_iter,
        random_state=random_state,
    ).fit(clustering_matrix)

    labels = kmedoids_clustering.labels_
    medoid_indices = list(kmedoids_clustering.medoid_indices_)
    cluster_centers = list(np.ravel(frames[medoid_indices]))
    cluster_centers_trajectories = list(np.ravel(trajectory_names[medoid_indices]))

    return labels, cluster_centers, medoid_indices, cluster_centers_trajectories



"""
    @release_date  : $release_date
    @version       : $release_version
    @author        : Christos Matsingos, Ka Fu Man

    This file is part of the TrIPP software
    (https://github.com/fornililab/TrIPP).
    Copyright (c) 2024 Christos Matsingos, Ka Fu Man and Arianna Fornili.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

import os
import multiprocessing as mp
from tripp._create_mda_universe_ import (
    create_mda_universe,
    create_propka_compatible_universe,
)
from tripp._pka_iterator_ import pka_iterator
from tripp._sort_pka_df_ import output_df
from datetime import datetime
from tripp._generate_trajectory_log_ import log_header, trajectory_log
import logging
import glob

class Trajectory:
    """
    Main class of TrIPP. Calling this class creates an iterable object of
    sliced trajectories that are then used with the run method to perform the
    analyses.

    Parameters
    ----------
    trajectory_file: str
        The path of the file containing the trajectory. The same formats
        permited by MDAnalysis can be used.
    topology_file: str
        The path of the file containing the topology. The same formats
        allowed by MDAnalysis can be used.
    output_directory: str
        The directory where output files will be saved.
    output_prefix: str
        The output file prefix.
    cpu_core_number: int, default=-1
        The number of cpu cores used for the calculation.
        If cpu_core_number=-1, all available cores are used.
    hetatm_resname: str, list, default=None
        PDB residue name(s) for non-protein molecules that we want to 
        be taken into account in the pKa calculation. Their record type
        will be set to 'HETATM'. See log for info if an error is raised.
    custom_terminal_oxygens: list, default=None
        PROPKA only recognizes C-terminal oxygen atoms named O and OXT. 
        If different, either provide a list of length 2 containing the 
        names of your C-terminal oxygen atoms, or set to None for correction 
        according to our pre-defined dictionary 
        (see _correction_dictionary_.py for the pre-defined names).
    custom_resname_correction: dict, default=None
        dictionary of custom protein residue names not included in
        the hard-coded TrIPP dictionary (tripp._correction_dictionary_.py). 
        Can be given as e.g. {'XXX':'ASP'}, where 'XXX' is the residue
        name in the PDB file and 'ASP' is the corresponding PROPKA name.

    """
    def __init__(
        self,
        topology_file,
        trajectory_file,
        output_directory,
        output_prefix,
        cpu_core_number=-1,
        hetatm_resname=None,
        custom_terminal_oxygens=None,
        custom_resname_correction=None,
    ):  
        self.output_directory = output_directory
        self.output_prefix = output_prefix
        if not os.path.isdir(output_directory):
            # Make directory if not present
            os.makedirs(output_directory) 
        else:
            # Remove files named .temp* in the directory before proceeding.
            [os.remove(file) for file in glob.glob(f'{output_directory}/.temp*')]
        if os.path.isfile(f'{output_directory}/{output_prefix}.log'):
            os.remove(f'{output_directory}/{output_prefix}.log')
            
        self.logger = logging.getLogger()
        self.logger.setLevel(logging.INFO)
        handler = logging.FileHandler(f'{output_directory}/{output_prefix}.log',
                                      'a')
        handler.setLevel(logging.INFO)
        formatter = logging.Formatter('%(message)s')
        handler.setFormatter(formatter)
        self.logger.addHandler(handler)
        propka_logger = logging.getLogger('propka') # Prevent loggers from PROPKA logging into our root logger.
        propka_logger.propagate = False 
        
        self.logger.info(log_header())
        self.logger.info('Trajectory class initialised:')
        
        self.trajectory_file = trajectory_file
        self.topology_file = topology_file

        if cpu_core_number == -1:
            self.cpu_core_number = os.cpu_count()

        else:
            self.cpu_core_number = cpu_core_number

        self.universe = create_mda_universe(
            topology_file=self.topology_file,
            trajectory_file=self.trajectory_file,
        )

        self.corrected_universe = create_propka_compatible_universe(
            self.universe,
            hetatm_resname,
            custom_terminal_oxygens,
            custom_resname_correction,
        )

        frames_nr = len(self.universe.trajectory)
        slices_nr = self.cpu_core_number
        slice_length = frames_nr // slices_nr
        remainder = frames_nr % slices_nr
        slices = []
        start_frame = 0

        for i in range(slices_nr):
            if i < remainder:
                end_frame = start_frame + slice_length + 1
            else:
                end_frame = start_frame + slice_length
            slices.append([start_frame, end_frame])
            start_frame = end_frame

        self.trajectory_slices = slices

    def run(
        self,
        extract_buriedness_data=True,
        mutation_selections=None,
        save_disulphide_pka=False,
        optargs=[],
    ):
        """
        Function to run PROPKA after initialising the Trajectory class

        Parameters
        ----------
        extract_buriedness_data: bool, default=True
            If set to True, both buried ratios and pKa values will be extracted.
            If set to False, only pKa values will be extracted.
        mutation_selections: str, default=None
            Peform pseudomutation of residues to alanine.
            Selection is based on MDAnalysis syntax. For multi-chain systems,
            please make sure you include the chainID in the selection. Double
            mutations can also be performed.
            e.g.: chainID A and resid 2 3
        save_disulphide_pka: bool, default=False
            If set to False, pKa and buried ratio values for cysteines 
            forming a disulphide bond (pKa set by PROPKA to 99.99) will 
            not be saved in the CSV files. 
        optargs: list of str, default=[]
            PROPKA predictions can be run with optional arguments
            (see https://propka.readthedocs.io/en/latest/command.html). 
            For example, if optargs is set to `["-k"]`, propka will run with the -k flag 
            (protons from the input file are kept).
        """
        start = datetime.today().strftime("%Y-%m-%d %H:%M:%S")
        
        pool = mp.Pool(self.cpu_core_number)
        # Create jobs
        jobs = []
        for trajectory_slice in self.trajectory_slices:
            # Create asynchronous jobs that will be submitted once a
            # processor is ready
            job = pool.apply_async(
                pka_iterator,
                args=(
                    trajectory_slice,
                    self.corrected_universe,
                    self.output_directory,
                    mutation_selections,
                    optargs),
            )
            jobs.append(job)
        # Submit jobs
        results = [job.get() for job in jobs]
        pool.close()
        pool.join()
        
        data, log_contents = zip(*results)
        self.data = data
        log_contents = list(filter(bool,log_contents))
        if log_contents:
            propka_warning_logger = logging.getLogger('proka_warning')
            propka_warning_logger.propagate = False
            propka_warning_handler = logging.FileHandler(f'{self.output_directory}/{self.output_prefix}_propka_warnings.log','w')
            propka_warning_handler.setLevel(logging.WARNING)
            propka_warning_handler.setFormatter(logging.Formatter('%(message)s'))
            propka_warning_logger.addHandler(propka_warning_handler)
            propka_warning_logger.warning(log_contents[0])
            propka_warning_handler.close()
            propka_warning_logger.removeHandler(propka_warning_handler)

        
        # Combine the temporary pka csv and sort it according to Time [ps] column
        disulphide_cys_col = output_df(
            output_directory=self.output_directory,
            output_prefix=self.output_prefix,
            data=data,
            save_disulphide_pka=save_disulphide_pka,
            extract_buriedness_data=extract_buriedness_data
        )

        end = datetime.today().strftime("%Y-%m-%d %H:%M:%S")

        # Log all parameter used for the run and also pKa statistic table.
        trajectory_log(
            self.output_directory,
            self.output_prefix,
            extract_buriedness_data,
            mutation_selections,
            disulphide_cys_col,
            optargs,
            self.cpu_core_number,
            self.trajectory_slices,
            start,
            end,
        )
        for handler in self.logger.handlers[:]:
            handler.close()
            self.logger.removeHandler(handler)


"""
    @release_date  : $release_date
    @version       : $release_version
    @author        : Christos Matsingos, Ka Fu Man 
    
    This file is part of the TrIPP software
    (https://github.com/fornililab/TrIPP).
    Copyright (c) 2024 Christos Matsingos, Ka Fu Man and Arianna Fornili.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
"""


""" 
Dictionary that contains the model pKa values of amino acids in solution used in 
PROPKA (https://github.com/jensengroup/propka/blob/27e0ad2f8d653404f57c78f1cae2ec32cb5adb68/propka/propka.cfg#L5-L13)
""" 


model_pka_values = {'ASP' : 3.8, 
                    'GLU' : 4.5, 
                    'CTR' : 3.2, 
                    'HIS' : 6.5, 
                    'NTR' : 8, 
                    'CYS' : 9, 
                    'TYR' : 10, 
                    'LYS' : 10.5, 
                    'ARG' : 12.5} 


"""
    @release_date  : $release_date
    @version       : $release_version
    @author        : Christos Matsingos, Ka Fu Man 
    
    This file is part of the TrIPP software
    (https://github.com/fornililab/TrIPP).
    Copyright (c) 2024 Christos Matsingos, Ka Fu Man and Arianna Fornili.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
from tripp._generate_trajectory_log_ import log_header

def generate_clustering_summary(trajectory_file, 
                                topology_file, 
                                pka_file, 
                                selections, 
                                include_distances, 
                                include_buriedness, 
                                clustering_method, 
                                automatic, 
                                silhouette_scores, 
                                n_components, 
                                cummulative_variance, 
                                buriedness_file):
    """
    Generates a summary of the clustering process.
    Parameters
    ----------
    trajectory_file: str or dict
        The path to the trajectory file or a dictionary with trajectory names as keys and file paths as values.
    topology_file: str
        The path to the topology file.
    pka_file: str or list
        The path to the pKa file or a list of paths if multiple pKa files are used.
    selections: list
        A list of residue selections used for clustering.
    include_distances: bool
        Whether distances between charge centers were included in the clustering.
    include_buriedness: bool
        Whether buriedness (buried ratio) values were included in the clustering.
    clustering_method: str
        The clustering method used, e.g., 'DBSCAN'.
    automatic: bool
        Whether automatic clustering (automatic scanning of clustering parameters) was performed.
    silhouette_scores: pd.DataFrame
        A DataFrame containing silhouette scores for different clustering parameters.
    n_components: int or None
        The number of principal components used for dimensionality reduction, or None if no reduction was done.
    cummulative_variance: float
        The cumulative variance explained by the principal components used.
    buriedness_file: str or list
        The path to the buriedness file or a list of paths if multiple buriedness files are used.
    Returns
    -------
    summary: str
        A formatted string summarizing the clustering process, including file information, residue selections, 
        whether distances and buriedness were included, clustering method, dimensionality reduction details, 
        automatic clustering parameters, and the best silhouette score parameters.
    """
    #information on files 
    if type(trajectory_file) == str: 
        trajectory_name = 'Unnnamed Trajectory' 
        if include_buriedness == False: 
            trajectory_file_summary= f'{trajectory_name} \nTrajectory file: {trajectory_file} \npKa file: {pka_file} \n\n' 
        elif include_buriedness == True: 
            trajectory_file_summary= f'{trajectory_name} \nTrajectory file: {trajectory_file} \npKa file: {pka_file}\nBuriedness file: {buriedness_file} \n\n' 

    else: 
        if include_buriedness == False: 
            trajectory_file_summary = '' 
            for trajectory_index, trajectory_name in enumerate(trajectory_file.keys()): 
                trajectory_file_summary+=f'{trajectory_name} \nTrajectory file: {trajectory_file[trajectory_name]} \npKa file: {pka_file[trajectory_index]} \n\n' 
        elif include_buriedness == True: 
            trajectory_file_summary = '' 
            for trajectory_index, trajectory_name in enumerate(trajectory_file.keys()): 
                trajectory_file_summary+=f'{trajectory_name} \nTrajectory file: {trajectory_file[trajectory_name]} \npKa file: {pka_file[trajectory_index]}\nBuriedness file: {buriedness_file[trajectory_index]} \n\n' 

    file_summary = f"""
Topology file: {topology_file} 

Trajectories: 

{trajectory_file_summary}
""" 
    
    #information on what residues were used for the clustering 
    residue_summary = f'Clustering was performed using {", ".join(selections[:-1])} and {selections[-1]}.' 

    #information on whether distances between charge centers were used for the clustering 
    if include_distances == True: 
        n = len(selections) 
        num_distances = int((n*(n-1))/2)
        if num_distances == 1: 
            include_distances_summary = f'In total {num_distances} distance between charge centers was included in the clustering.' 
        else: 
            include_distances_summary = f'In total {num_distances} distances between charge centers were included in the clustering.' 
    
    elif include_distances == False: 
        include_distances_summary = f'No distances between charge centers were included in the clustering.' 
    
    #information on whether buriedness measures were used for the clustering 
    if include_buriedness== True: 
        n = len(selections) 
        if n == 1: 
            include_buriedness_summary = f'In total {n} buriedness measure was included in the clustering.' 
        else: 
            include_buriedness_summary = f'In total {n} buriedness measures were included in the clustering.' 
    
    elif include_buriedness == False: 
        include_buriedness_summary = f'No buriedness measures were included in the clustering.' 
    
    #information on clustering method 
    clustering_method_summary = f'The {clustering_method} method was used for the clustering.' 

    #information on dimensionality reduction 
    if n_components == None: 
        dimensionality_reduction_summary = 'No dimensionality reduction was performed.' 
    
    else: 
        dimensionality_reduction_summary = f'Dimensionality reduction was performed using PCA.\nIn total {n_components} principal components were used with a cummulative variance of {cummulative_variance}%.' 
    
    #information on the silhouette score 
    best_index = silhouette_scores['Average silhouette score'].idxmax() 
    best_params = '' 
    for params in silhouette_scores.columns: 
        best_params+=f'{params}: {silhouette_scores[params].iloc[best_index]}\n' 
    
    best_params_summary = f'Clustering was performed using the following parameters: \n{best_params} '
    
    #information on all parameters tested 
    if automatic == True: 
        automatic_clustering_summary = f'Automatic clustering was selected. \nThe following parameters were tested: \n\n{silhouette_scores.to_markdown(index=False)}' 
    
    elif automatic == False: 
        automatic_clustering_summary = 'No automatic clustering was performed' 
    
    summary = f"""{log_header()} 

{file_summary} 
{residue_summary} 
{include_distances_summary} 
{include_buriedness_summary} 
{clustering_method_summary} 
{dimensionality_reduction_summary} 

{automatic_clustering_summary} 

{best_params_summary} 

-----------------------------------------------------------------

Cluster details: 
""" 
    
    return summary 


"""
    @release_date  : $release_date
    @version       : $release_version
    @author        : Christos Matsingos, Ka Fu Man 
    
    This file is part of the TrIPP software
    (https://github.com/fornililab/TrIPP).
    Copyright (c) 2024 Christos Matsingos, Ka Fu Man and Arianna Fornili.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np 

def sort_clusters(labels, cluster_centers, cluster_center_indices, cluster_centers_trajectories): 
    """
    Sorts clusters based on their population and reassigns labels.
    Parameters
    ----------
    labels : np.ndarray
        An array of cluster labels for each point in the clustering matrix                 (feature matrix).
    cluster_centers: list
        Indices of the cluster centers within their individual trajectories (local indices). 
        Indices start from 0 for each trajectory.
    cluster_center_indices: list
        Indices of the cluster centers within the full feature matrix (global indices).
    cluster_centers_trajectories: list
        Trajectory names of the cluster centers.
    """
    clusters = np.unique(labels) 
    clusters = clusters[clusters != -1] 
    labels_without_outliers = labels[labels != -1]
    
    frequency = np.bincount(labels_without_outliers, minlength=len(clusters))  
    
    new_labels = np.zeros_like(labels, dtype=int) 
    new_labels[labels == -1] = -1 
    
    for i in range(len(clusters)): 
        index_biggest = np.argmax(frequency) 
        biggest_cluster = clusters[index_biggest] 
        labels_mask = labels == biggest_cluster 
        new_labels[labels_mask] = i 
        frequency[index_biggest] = 0 
        
    
    new_cluster_centers = np.zeros_like(clusters, dtype=int) 
    new_cluster_center_indices = np.zeros_like(clusters, dtype=int) 
    new_cluster_centers_trajectories = np.zeros_like(clusters, dtype=object) 

    for i in range(len(clusters)): 
        cluster_index = cluster_center_indices[i] 
        new_label = new_labels[cluster_index] 
        new_cluster_center_indices[new_label] = cluster_center_indices[i] 
        new_cluster_centers[new_label] = cluster_centers[i] 
        new_cluster_centers_trajectories[new_label] = cluster_centers_trajectories[i] 
    
    return new_labels, new_cluster_centers, new_cluster_center_indices, new_cluster_centers_trajectories 


"""
    @release_date  : $release_date
    @version       : $release_version
    @author        : Christos Matsingos, Ka Fu Man 
    
    This file is part of the TrIPP software
    (https://github.com/fornililab/TrIPP).
    Copyright (c) 2024 Christos Matsingos, Ka Fu Man and Arianna Fornili.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

def mutate(universe, mutation_selections, temp_name): 
    """ 
    Function that deletes all atoms of a residue except for a methyl group. 
    The residue is then renamed to alanine. 
    Parameters
    ----------
    universe: MDAnalysis.universe
        MDAnalysis universe. The universe after create_propka_compatible_universe
        modifications should be used here.
    mutation_selections: str
        A string (MDAnalysis selection syntax) to select the 
        residues to be pseudo-mutated.
    temp_name: str
        The name of the temporary PDB file to be created with the mutated residue.
    Exceptions
    ----------
    Exception
        If the residue type is GLY, an exception is raised.
    """ 
    replace_name = ' '.join(['N','HN','H','CA','HA','CB','O','C'])
    mutation_ag = universe.select_atoms(mutation_selections)
    if (mutation_ag.residues.resnames).any() == 'GLY':
        raise Exception('GLY cannot be mutated to Ala in the current implementation.')
    mutation_ag.residues.resnames = 'ALA'
    # Selecting all but not the mutation_selection, and also the mutation_selection but only those of replace_name.
    mutation_ag = universe.select_atoms(f"(all and not ({mutation_selections})) or ({mutation_selections} and name {replace_name})")
    mutation_ag.write(f'{temp_name}.pdb')



"""
    @release_date  : $release_date
    @version       : $release_version
    @author        : Christos Matsingos, Ka Fu Man

    This file is part of the TrIPP software
    (https://github.com/fornililab/TrIPP).
    Copyright (c) 2024 Christos Matsingos, Ka Fu Man and Arianna Fornili.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np


def calculate_charge_center_distances(positions):
    """
    Function that calculates the distances between the charge centers
    of n residues. The function takes as input a n x 3 matrix
    containing the coordinates of the charge centers
    and calculates n(n-1)/2 distances.
    
    Parameters
    ----------
    positions : np.ndarray
        An n x 3 numpy array containing the coordinates of the charge centers.
    Returns
    -------
    distance_array : np.ndarray
        A 1D numpy array containing the distances between each pair of charge centers
    """

    distances = []
    for i in range(len(positions)):
        center_i = positions[i]
        for j in range(i + 1, len(positions)):
            center_j = positions[j]
            sqr_sum = np.sum(np.square(center_i - center_j))
            d = np.sqrt(sqr_sum)
            distances.append(d)

    distance_array = np.array(distances)

    return distance_array



"""
    @release_date  : $release_date
    @version       : $release_version
    @author        : Christos Matsingos, Ka Fu Man 
    
    This file is part of the TrIPP software
    (https://github.com/fornililab/TrIPP).
    Copyright (c) 2024 Christos Matsingos, Ka Fu Man and Arianna Fornili.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
"""


import subprocess
import os
def gen_pymol_template(tempfactors_topology_file,
                       pymol_path, 
                       pse_output_filename, 
                       values_df, 
                       lower_limit, 
                       upper_limit, 
                       color_palette): 

    """
    Function to colour-map residue properties on a reference structure using PyMOL.
    The values of the parameters are determined by gen_pse()
    from Visualization class.

    Parameters:
    tempfactors_topology_file: str
    PDB file with the input topology and tempfactors assigned.
    
    pymol_path: str
    Path to PyMOL. The script will spawn a subprocess shell to run a 
    python script in PyMOL.

    pse_output_filename: str
    Name of the PyMOL session file (pse) produced as output. The name is 
    a combination of pse_output_prefix and the colouring method.

    values_df: Pandas DataFrame
    DataFrame with two columns, one for the residue identifier and one for
    the value to be mapped (pKa, pKa difference or correlation)

    lower_limit: int or float
    Lower bound of the color scale in the PyMOL session. Residues with values
    below this threshold are assigned the gradient’s minimum color.

    upper_limit: int or float
    Upper bound of the color scale in the PyMOL session. Residues with values
    above this threshold are assigned the gradient’s maximum color.

    color_palette: str
    Colour palette. The default is set to 'red_white_blue'. See PyMOL spectrum
    for allowed colour palettes (a three-colour palette is recommended).
    """
    with open('.pymol_template.py','a') as output:
        output.write(f"""cmd.load('{tempfactors_topology_file}', 'protein_str')
cmd.show("cartoon", 'protein_str')
cmd.color("white", "protein_str")\n""")
    names = []
    residue_identifiers, values = (columns for _,columns in values_df.items()) 
    for residue_identifier, value in zip(residue_identifiers,values):
        residue = residue_identifier.split(':')[0]
        chain = residue_identifier.split(':')[-1]
        rounded_value = round(value,2)
        if 'N+' in residue:
            resid = residue[2:]
            name = f'NTR{resid}_{chain}'
            selection = f'(bb. and not elem O and not elem C and byres protein_str and chain {chain} and resi {resid}) extend 1 and not elem C'
            label_sel = f'{name} and bb. and elem N'
        elif 'C-' in residue:
            resid = residue[2:]
            name = f'CTR{resid}_{chain}'
            selection = f'((bb. and byres protein_str and chain {chain} and resi {resid}) and elem C and not name CA) extend 1 and not name CA'
            label_sel = f'{name} and bb. and elem C and not name CA'
        else:
            name = f'{residue}_{chain}'
            resid = residue[3:]
            selection = f'((byres protein_str and chain {chain} and resi {resid})&(sc.|(n. CA|n. N&r. PRO))) and not name H1+H2+H3 and not (protein_str and chain {chain} and resi {resid} & name C extend 1 &! name CA)'
            label_sel = f'{name} and name CB'
        names.append(name)
        with open('.pymol_template.py', 'a') as output:
            output.write(f"""cmd.create('{name}', '{selection}') 
cmd.show('licorice', '{name}') 
cmd.spectrum('b','{color_palette}','{name}',{lower_limit},{upper_limit})
cmd.label('{label_sel}','{rounded_value}')\n""")
    
    sorted_residues = ' '.join(sorted(names, key=lambda x: (x[-1], int(x[3:-2]))))
    with open('.pymol_template.py', 'a') as output:
        output.write(f"""cmd.order('{sorted_residues}')
cmd.ramp_new('colorbar', 'none', [{lower_limit}, ({lower_limit} + {upper_limit})/2, {upper_limit}], {color_palette.split('_')})
cmd.set('label_size','-2')
cmd.set('label_position','(1.2,1.2,1.2)')
cmd.orient('protein_str')
cmd.bg_color('white')
cmd.set('orthoscopic')
cmd.set('depth_cue',0)
cmd.save('{pse_output_filename}')
cmd.quit()\n""")
    subprocess.run([f'{pymol_path} -c .pymol_template.py'],shell=True)
    os.remove('.pymol_template.py')


"""
    @release_date  : $release_date
    @version       : $release_version
    @author        : Christos Matsingos, Ka Fu Man 
    
    This file is part of the TrIPP software
    (https://github.com/fornililab/TrIPP).
    Copyright (c) 2024 Christos Matsingos, Ka Fu Man and Arianna Fornili.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

import MDAnalysis as mda
import pandas as pd
from tripp._create_clustering_matrix_ import create_clustering_matrix
from tripp._clustering_kmedoids_ import kmedoids_clustering
from tripp._write_clustering_info_ import write_clustering_info
from sklearn.metrics import silhouette_score
import numpy as np
from tripp._calculate_rmsd_matrix_ import calculate_rmsd_matrix
from tripp._clustering_greedy_ import greedy_clustering
from tripp._clustering_dbscan_ import dbscan_clustering
from tripp._clustering_hdbscan_ import hdbscan_clustering
from tripp._pca_ import pca
from tripp._generate_clustering_summary_ import generate_clustering_summary
from tripp._determine_cluster_population_ratio_ import (
    determine_cluster_population_ratio,
)
import logging

class Clustering:
    """
    This class provides a way to extract representative structures from a trajectory
    using the clustering methods: kmedoids, greedy, or dbscan methods. Clustering
    is carried out using the pKa values and (optional) the buried ratio and 
    the relative position of selected residues as features.

    Parameters
    ----------
    topology_file: str
        Path to the topology file. The same formats allowed by MDAnalysis can be used.
    trajectory_file: str or dict
        When str, it is the path of the file containing the trajectory. The same
        formats allowed by MDAnalysis can be used. When dict, the clustering is performed using
        multiple trajectories specified in a dictionary, where trajectory names are
        used as key and file paths as values, e.g.: {'MD1' : 'file1 path', 'MD2' : 'file2 path', ...}.
        The same topology file is used for all trajectories.
    pka_file: str or list
        When str, it is the path of the CSV file with the pka values generated by the
        Trajectory class using trajectory_file and topology_file. 
        When list, it is a list of paths of the CSV pKa files for the
        trajectories in trajectory_file. In this case, the file paths need to be given in the 
        same order as the correspnding trajectories in trajectory_file.
    selections: list of str
        List of residues used for the clustering (selected with MDAnalysis selection syntax)
        For multi-chain systems, your topology and selection must include chain IDs.
    output_directory: str
        Directory where the output files will be saved.
    output_prefix: str
        Output file prefix
    buriedness_file: str or list, default=None
        When str, it is the path of the CSV file with the buried ratio values generated by the
        Trajectory class using trajectory_file and topology_file. 
        When list, it is a list of paths of the CSV buried ratio files for the
        trajectories in trajectory_file. In this case, the file paths need to be given in the 
        same order as the correspnding trajectories in trajectory_file.
        If None, the buried ratio values are not used for clustering.
    include_distances: bool, default=False
        If True, all the possible pairs of distances between charge centers in the selected 
        residues are used as additional features for the clustering alongside the pKa values.
    include_buriedness: bool, default=False
        If True, the buried ratio of the selected residues are used as additional features
        for the clustering alongside the pKa values.
    dimensionality_reduction: bool, befault=False
        If True, dimensionality reduction is performed on the input features using PCA. 
        The minimum number of principal components that explain at least 90% of the 
        total variance is used.
    """
    
    def __init__(
        self,
        topology_file,
        trajectory_file,
        pka_file,
        selections,
        output_directory,
        output_prefix,
        buriedness_file=None,
        include_distances=False,
        include_buriedness=False,
        dimensionality_reduction=False,
    ):
        self.logger = logging.getLogger()
        if self.logger.hasHandlers:
            for handler in self.logger.handlers[:]:
                self.logger.removeHandler(handler)
        logging.Formatter('%(message)s')
        self.logger.setLevel(logging.WARNING)
        self.topology_file = topology_file
        self.trajectory_file = trajectory_file
        self.pka_file = pka_file
        self.selections = selections
        self.output_directory = output_directory
        self.output_prefix = output_prefix
        self.buriedness_file = buriedness_file
        self.include_buriedness = include_buriedness
        self.include_distances = include_distances
        self._labels = None
        self._cluster_center_indices = None

        def make_pka_or_buriedness_df(file):
            if isinstance(file, str):
                df = pd.read_csv(file, index_col="Time [ps]")
                df["Trajectories"] = "Unnamed trajectory"

            elif isinstance(file, list):
                df_list = []
                for index, f in enumerate(file):
                    df = pd.read_csv(f, index_col="Time [ps]")
                    df["Trajectories"] = list(self.trajectory_file.keys())[index]
                    df_list.append(df)
                df = pd.concat(df_list)

            elif file is None:
                df = None

            return df

        self.pka_df = make_pka_or_buriedness_df(pka_file)
        self.buriedness_df = make_pka_or_buriedness_df(buriedness_file)

        (
            self.clustering_matrix,
            self.times,
            self.frames,
            self.trajectory_names,
            self.trajectory_dict,
        ) = create_clustering_matrix(
            self.trajectory_file,
            self.topology_file,
            self.pka_df,
            self.selections,
            self.include_distances,
            self.buriedness_df,
            self.include_buriedness,
        )

        if dimensionality_reduction:
            self.n_components, self.cummulative_variance, self.clustering_matrix = pca(
                self.clustering_matrix
            )

        elif not dimensionality_reduction:
            self.cummulative_variance = None
            self.n_components = None

    def kmedoids(
        self,
        automatic=False,
        n_clusters=None,
        metric="euclidean",
        method="alternate",
        init="k-medoids++",
        max_iter=300,
        random_state=None,
        max_clusters=20,
    ):
        """
        Performs clustering with K-Medoids.
        Parameters
        ----------
        automatic: bool, default=False
            If True, clustering is run with n_clusters ranging from 2 to 
            max_clusters. The silhouette score is determined 
            at each iteration and the best number of clusters is chosen 
            based on the highest average silhouette score.

        n_clusters: int, default=8
            Number of output clusters. Ignored if 
            automatic=True.

        metric, method, init, max_iter, random_state as found in
        sklearn_extra.cluster.KMedoids (https://scikit-learn-extra.readthedocs.io/en/stable/generated/sklearn_extra.cluster.KMedoids.html).

        max_clusters: int, default=20.
            Maximum number of clusters allowed when automatic=True.
        """

        clustering_method = "KMedoids"

        if automatic == False:
            if not n_clusters:
                raise Exception("n_clusters must be specified when automatic=False.") 
            labels, cluster_centers, medoid_indices, cluster_centers_trajectories = (
                kmedoids_clustering(
                    n_clusters=n_clusters,
                    metric=metric,
                    method=method,
                    init=init,
                    max_iter=max_iter,
                    random_state=random_state,
                    clustering_matrix=self.clustering_matrix,
                    frames=self.frames,
                    trajectory_names=self.trajectory_names,
                )
            )
            sil_score = round(silhouette_score(self.clustering_matrix, labels), 4)
            print(
                f"Clustering with {n_clusters} clusters produces an average silhouette score of {sil_score}."
            )
            silhouette_scores = pd.DataFrame(
                {
                    "Number of clusters": [n_clusters],
                    "Average silhouette score": [sil_score],
                }
            )

        elif automatic == True:
            sil_scores = []
            cluster_nums = []
            for clusters in range(2, max_clusters + 1):
                (
                    labels,
                    cluster_centers,
                    medoid_indices,
                    cluster_centers_trajectories,
                ) = kmedoids_clustering(
                    n_clusters=clusters,
                    metric=metric,
                    method=method,
                    init=init,
                    max_iter=max_iter,
                    random_state=random_state,
                    clustering_matrix=self.clustering_matrix,
                    frames=self.frames,
                    trajectory_names=self.trajectory_names,
                )
                try:
                    sil_score = round(silhouette_score(self.clustering_matrix, labels), 4)
                    sil_scores.append(sil_score)
                    cluster_nums.append(clusters)
                    print(
                        f"Clustering with {clusters} clusters produces an average silhouette score of {sil_score}."
                    )
                except Exception as e:
                    print(f'Error found when evaluating silhouette score with {clusters} clusters so skipping this: {e}.')
                    continue
            sil_scores = np.array(sil_scores)
            best_cluster_n_index = np.argmax(sil_scores)
            sil_score = sil_scores[best_cluster_n_index]
            n_clusters = cluster_nums[best_cluster_n_index]
            labels, cluster_centers, medoid_indices, cluster_centers_trajectories = (
                kmedoids_clustering(
                    n_clusters=n_clusters,
                    metric=metric,
                    method=method,
                    init=init,
                    max_iter=max_iter,
                    random_state=random_state,
                    clustering_matrix=self.clustering_matrix,
                    frames=self.frames,
                    trajectory_names=self.trajectory_names,
                )
            )
            print(
                f"Best number of clusters identified at {n_clusters} with an average silhouette score of {sil_score}."
            )

            silhouette_scores = pd.DataFrame(
                {
                    "Number of clusters": cluster_nums,
                    "Average silhouette score": sil_scores,
                }
            )
        
        summary = generate_clustering_summary(
            trajectory_file=self.trajectory_file,
            topology_file=self.topology_file,
            pka_file=self.pka_file,
            selections=self.selections,
            include_distances=self.include_distances,
            include_buriedness=self.include_buriedness,
            clustering_method=clustering_method,
            automatic=automatic,
            silhouette_scores=silhouette_scores,
            n_components=self.n_components,
            cummulative_variance=self.cummulative_variance,
            buriedness_file=self.buriedness_file,
        )
        self._labels, self._cluster_centers, self._cluster_center_indices, self._cluster_centers_trajectories  = (
        write_clustering_info(
            summary=summary,
            trajectory_dict=self.trajectory_dict,
            pka_df=self.pka_df,
            times=self.times,
            frames=self.frames,
            trajectory_names=self.trajectory_names,
            labels=labels,
            cluster_centers=cluster_centers,
            cluster_center_indices=medoid_indices,
            cluster_centers_trajectories=cluster_centers_trajectories,
            output_directory=self.output_directory,
            output_prefix=self.output_prefix,
            clustering_method=clustering_method,
            )
        )

    def greedy(
        self,
        automatic=False,
        max_cluster_population=0.95,
        cutoff=0.1,
        max_clusters=20,
        max_cutoffs=20,
    ):
        """
        This function implements the greedy clustering method
        as described in Micheletti et al.
        (DOI: 10.1002/1097-0134(20000901)40:4<662::aid-prot90>3.0.co;2-f).

        automatic: bool, default=False
            If True, clustering is run with a range of distance cutoffs defined by 
            np.arange(step, max_rmsd + step, step), where step = max_rmsd / max_cutoffs
            and max_rmsd is the maximum distance.
            The silhouette score is determined for each cutoff and the best cutoff is chosen 
            based on the highest silhouette score. Cutoffs that produce more than the 
            allowed maximum number of clusters (defined by max_clusters) or that produce 
            only one cluster are excluded.
        
        max_cluster_population: float, default=0.95
            Used when automatic=True. The maximum allowed population value for the most
            populated cluster. If the population of the most populated cluster exceeds 
            this value, the clustering is skipped.
        
        cutoff: float, default=0.1
            Distance cutoff used for clustering. Ignored if automatic=True.

        max_clusters: int, default=20
            Maximum number of clusters allowed when automatic=True.

        max_cutoffs: int, default=20
            Maximum number of distance cutoffs used when automatic=True.
        """
        clustering_method = "greedy"

        rmsd_matrix = calculate_rmsd_matrix(self.clustering_matrix, self.frames)

        if automatic == False:
            # Warning message for default values when automatic=False
            if cutoff == 0.1:
                print("Warning: 'automatic' is set to False and the default value for 'cutoff' (0.1) will be used. Consider setting this parameter explicitly for better results.") 
            (
                labels,
                cluster_centers,
                cluster_center_indices,
                cluster_centers_trajectories,
            ) = greedy_clustering(
                cutoff=cutoff,
                rmsd_matrix=rmsd_matrix,
                frames=self.frames,
                trajectory_names=self.trajectory_names,
            )
            sil_score = round(silhouette_score(self.clustering_matrix, labels), 4)
            print(
                f"Clustering with a cutoff of {cutoff} produces {len(set(labels))} clusters with an average silhouette score of {sil_score}."
            )
            silhouette_scores = pd.DataFrame(
                {
                    "Number of clusters": [len(set(labels))],
                    "RMSD cutoff": [cutoff],
                    "Average silhouette score": [sil_score],
                }
            )

        elif automatic == True:
            sil_scores = []
            max_rmsd = np.max(rmsd_matrix)
            step = max_rmsd / max_cutoffs
            cutoffs = []
            cluster_nums = []
            for cutoff_i in np.arange(step, max_rmsd + step, step):
                cutoff_i = round(cutoff_i, 4)
                (
                    labels,
                    cluster_centers,
                    cluster_center_indices,
                    cluster_centers_trajectories,
                ) = greedy_clustering(
                    cutoff=cutoff_i,
                    rmsd_matrix=rmsd_matrix,
                    frames=self.frames,
                    trajectory_names=self.trajectory_names,
                )

                if len(labels) == len(set(labels)):
                    print(
                        f"A distance cutoff of {cutoff_i} produces no clusters. Moving on to next cutoff"
                    )

                elif len(set(labels)) == 1:
                    print(
                        f"A distance cutoff of {cutoff_i} produces only one cluster. Moving on to next cutoff."
                    )

                elif len(set(labels)) > max_clusters:
                    print(
                        f"A distance cutoff of {cutoff_i} produces more than {max_clusters} clusters. Moving on to next cutoff."
                    )

                elif (
                    determine_cluster_population_ratio(
                        labels=labels,
                        max_cluster_population=max_cluster_population,
                        clustering_method=clustering_method,
                    )
                    == True
                ):
                    print(
                        f"Clustering with a cutoff of {cutoff_i} results in the most populated cluster containing more than {max_cluster_population*100}% of frames. Moving on to next cutoff."
                    )

                else:
                    cutoffs.append(cutoff_i)
                    try: 
                        sil_score = round(silhouette_score(self.clustering_matrix, labels), 4)
                        sil_scores.append(sil_score)
                        cluster_nums.append(len(set(labels)))
                        print(f"Clustering with a cutoff of {cutoff_i} produces {len(set(labels))} clusters with an average silhouette score of {sil_score}.")
                    except Exception as e:
                        print(f'Error found when evaluating silhouette score with distance cutoff {cutoff_i} so skipping this: {e}.')
                        continue
                        

            sil_scores = np.array(sil_scores)
            best_cutoff_index = np.argmax(sil_scores)
            cutoff = cutoffs[best_cutoff_index]
            sil_score = sil_scores[best_cutoff_index]
            print(
                f"Best cutoff identified at {cutoff} with an average silhouette score of {sil_score}."
            )
            (
                labels,
                cluster_centers,
                cluster_center_indices,
                cluster_centers_trajectories,
            ) = greedy_clustering(
                cutoff=cutoff,
                rmsd_matrix=rmsd_matrix,
                frames=self.frames,
                trajectory_names=self.trajectory_names,
            )

            silhouette_scores = pd.DataFrame(
                {
                    "Number of clusters": cluster_nums,
                    "Distance cutoff": cutoffs,
                    "Average silhouette score": sil_scores,
                }
            )
        
        summary = generate_clustering_summary(
            trajectory_file=self.trajectory_file,
            topology_file=self.topology_file,
            pka_file=self.pka_file,
            selections=self.selections,
            include_distances=self.include_distances,
            include_buriedness=self.include_buriedness,
            clustering_method=clustering_method,
            automatic=automatic,
            silhouette_scores=silhouette_scores,
            n_components=self.n_components,
            cummulative_variance=self.cummulative_variance,
            buriedness_file=self.buriedness_file,
        )
        self._labels, self._cluster_centers, self._cluster_center_indices, self._cluster_centers_trajectories  = (
            write_clustering_info(
            summary=summary,
            trajectory_dict=self.trajectory_dict,
            pka_df=self.pka_df,
            times=self.times,
            frames=self.frames,
            trajectory_names=self.trajectory_names,
            labels=labels,
            cluster_centers=cluster_centers,
            cluster_center_indices=cluster_center_indices,
            cluster_centers_trajectories=cluster_centers_trajectories,
            output_directory=self.output_directory,
            output_prefix=self.output_prefix,
            clustering_method=clustering_method,
            )
        )

    def dbscan(
        self,
        automatic=False,
        eps_range=(0.005, 0.505, 0.005),
        min_samples_range=(2, 11, 1),
        max_cluster_population=0.95,
        max_clusters=20,
        eps=0.5,
        min_samples=5,
        metric="euclidean",
        metric_params=None,
        algorithm="auto",
        leaf_size=30,
        p=None,
        n_jobs=None,
    ):
        """
        This function implements the DBSCAN method to do the clustering.

        automatic: bool, default=False
            If True, the clustering is run using multiple combinations of eps and 
            min_samples values. The silhouette score is determined for each combination and
            the combination with the highest silhouette score is selected. Combinations 
            that produce more than max_clusters or only one cluster are excluded.

        eps_range: tuple, default=(0.005, 0.505, 0.005)
            The range of epsilon values screened in the automatic grid search. Interpreted as 
            (start, stop, step), values are generated with np.arange(start, stop, step).

        min_samples_range: tuple, default=(2, 11, 1)
            The range of min_samples values screened in the automatic grid search. Interpreted as 
            (start, stop, step), values are generated with np.arange(start, stop, step).
        
        max_cluster_population: float, default=0.95
            Used when automatic=True. The maximum allowed population value for the most
            populated cluster. If the population of the most populated cluster exceeds 
            this value, the clustering is skipped.
            
        max_clusters: int, default=20
            Maximum number of clusters allowed when automatic=True.

        eps, min_samples, metric, metric_params, algorithm,
        leaf_size, p, and n_jobs as found in
        sklearn.cluster.DBSCAN (https://scikit-learn.org/stable/modules/generated/sklearn.cluster.DBSCAN.html).
        """

        clustering_method = "DBSCAN"

        if automatic == False:
            if eps == 0.5 and min_samples == 5:
                print("Warning: 'automatic' is set to False and the default values for 'eps' (0.5) and 'min_samples' (5) will be used. Consider setting these parameters explicitly for better results.") 
            (
                labels,
                cluster_centers,
                cluster_center_indices,
                cluster_centers_trajectories,
            ) = dbscan_clustering(
                eps=eps,
                min_samples=min_samples,
                metric=metric,
                metric_params=metric_params,
                algorithm=algorithm,
                leaf_size=leaf_size,
                p=p,
                n_jobs=n_jobs,
                clustering_matrix=self.clustering_matrix,
                frames=self.frames,
                find_centroid=True,
                trajectory_names=self.trajectory_names,
            )
            sil_score = round(silhouette_score(self.clustering_matrix, labels), 4)
            print(
                f"Clustering with parameters eps={eps} and min_samples={min_samples} produces {len(set(labels))} clusters with an average silhouette score of {sil_score}."
            )
            silhouette_scores = pd.DataFrame(
                {
                    "Number of clusters": [len(set(labels))],
                    "Epsilon": [eps],
                    "Minimum samples": [min_samples],
                    "Average silhouette score": [sil_score],
                }
            )

        elif automatic == True:
            sil_scores = []
            params = []
            cluster_nums = []
            for eps in np.arange(eps_range[0], eps_range[1], eps_range[2]):
                eps = round(eps, 4)
                for min_samples in np.arange(
                    min_samples_range[0], min_samples_range[1], min_samples_range[2]
                ):
                    labels = dbscan_clustering(
                        eps=eps,
                        min_samples=min_samples,
                        metric=metric,
                        metric_params=metric_params,
                        algorithm=algorithm,
                        leaf_size=leaf_size,
                        p=p,
                        n_jobs=n_jobs,
                        clustering_matrix=self.clustering_matrix,
                        frames=self.frames,
                        find_centroid=False,
                        trajectory_names=self.trajectory_names,
                    )

                    if len(set(labels)) == len(labels):
                        print(
                            f"Clustering with parameters eps={eps} and min_samples={min_samples} produces no clusters. Moving on to next set of parameters"
                        )

                    elif len(set(labels)) <= 2:
                        print(
                            f"Clustering with parameters eps={eps} and min_samples={min_samples} produces only one cluster. Moving on to next set of parameters"
                        )

                    elif len(set(labels)) > max_clusters:
                        print(
                            f"Clustering with parameters eps={eps} and min_samples={min_samples} produces more than {max_clusters} clusters. Moving on to next set of parameters"
                        )

                    elif (
                        determine_cluster_population_ratio(
                            labels=labels,
                            max_cluster_population=max_cluster_population,
                            clustering_method=clustering_method,
                        )
                        == True
                    ):
                        print(
                            f"Clustering with parameters eps={eps} and min_samples={min_samples} results in the most populated cluster containing more than {max_cluster_population*100}% of frames. Moving on to next set of parameters."
                        )

                    else:
                        try:
                            sil_score = round(
                                silhouette_score(self.clustering_matrix, labels), 4
                            )
                            sil_scores.append(sil_score)
                            cluster_nums.append(len(set(labels)))
                            params.append([eps, min_samples])
                            print(
                                f"Clustering with parameters eps={eps} and min_samples={min_samples} produces {len(set(labels))} clusters with an average silhouette score of {sil_score}."
                            )
                        except Exception as e:
                            print(f'Error found when evaluating silhouette score with eps={eps} and min_samples={min_samples} so skipping this: {e}.')
                            continue

            sil_scores = np.array(sil_scores)
            best_params_index = np.argmax(sil_scores)
            best_eps, best_min_samples = tuple(params[best_params_index])
            sil_score = sil_scores[best_params_index]
            print(
                f"Best clustering parameters identified at eps={best_eps} and min_samples={best_min_samples} with an average silhouette score of {sil_score}."
            )
            (
                labels,
                cluster_centers,
                cluster_center_indices,
                cluster_centers_trajectories,
            ) = dbscan_clustering(
                eps=best_eps,
                min_samples=best_min_samples,
                metric=metric,
                metric_params=metric_params,
                algorithm=algorithm,
                leaf_size=leaf_size,
                p=p,
                n_jobs=n_jobs,
                clustering_matrix=self.clustering_matrix,
                frames=self.frames,
                find_centroid=True,
                trajectory_names=self.trajectory_names,
            )
            params = np.array(params)

            silhouette_scores = pd.DataFrame(
                {
                    "Number of clusters": cluster_nums,
                    "Epsilon": params[:, 0],
                    "Minimum samples": params[:, 1],
                    "Average silhouette score": sil_scores,
                }
            )
        self._labels = labels
        self._cluster_center_indices = cluster_center_indices
        
        summary = generate_clustering_summary(
            trajectory_file=self.trajectory_file,
            topology_file=self.topology_file,
            pka_file=self.pka_file,
            selections=self.selections,
            include_distances=self.include_distances,
            include_buriedness=self.include_buriedness,
            clustering_method=clustering_method,
            automatic=automatic,
            silhouette_scores=silhouette_scores,
            n_components=self.n_components,
            cummulative_variance=self.cummulative_variance,
            buriedness_file=self.buriedness_file,
        )
        self._labels, self._cluster_centers, self._cluster_center_indices, self._cluster_centers_trajectories  = (
            write_clustering_info(
            summary=summary,
            trajectory_dict=self.trajectory_dict,
            pka_df=self.pka_df,
            times=self.times,
            frames=self.frames,
            trajectory_names=self.trajectory_names,
            labels=labels,
            cluster_centers=cluster_centers,
            cluster_center_indices=cluster_center_indices,
            cluster_centers_trajectories=cluster_centers_trajectories,
            output_directory=self.output_directory,
            output_prefix=self.output_prefix,
            clustering_method=clustering_method,
            )
        )
        
    def hdbscan(
        self,
        automatic=False,
        min_cluster_size_range=(20, 80, 5),
        min_samples_range=(20, 100, 5),
        max_cluster_population=0.95,
        max_clusters=20,
        min_cluster_size=5,
        min_samples=None,
        **kwargs):
        """
        This function implements the HDBSCAN method to do the clustering.

        automatic: bool, default=False
            If True, the clustering is run using various combinations of min_cluster_size and
            min_samples values. The silhouette score is determined for each combination and
            the combination with the highest silhouette score is selected. Combinations 
            that produce more than max_clusters or only one cluster are excluded.
        
        max_cluster_population: float, default=0.95
            Used when automatic=True. The maximum allowed population value for the most
            populated cluster. If the population of the most populated cluster exceeds 
            this value, the clustering is skipped.

        min_cluster_size_range: tuple, default=(20, 80, 5)
            The range of min_cluster_size screened in the automatic grid search. Interpreted as 
            (start, stop, step), values are generated with np.arange(start, stop, step).

        min_samples_range: tuple, default=(20, 100, 5)
            The range of min_samples value screened in the automatic grid search. Interpreted as 
            (start, stop, step), values are generated with np.arange(start, stop, step).

        max_clusters: int, default=2
            Maximum number of clusters allowed when automatic=True.

        kwargs:
            Other arguments allowed in HDBSCAN can be provided.
            (https://scikit-learn.org/stable/modules/generated/sklearn.cluster.HDBSCAN.html)
        """

        clustering_method = "HDBSCAN"

        if automatic == False:
            if max_cluster_population == 0.95 and min_cluster_size == 5 and not min_samples:
                print("Warning: 'automatic' is set to False and the default values for 'max_cluster_population' (0.95), 'min_cluster_size' (5), and 'min_samples' (None) will be used. Consider setting these parameters explicitly for better results.") 
            (
                labels,
                cluster_centers,
                cluster_center_indices,
                cluster_centers_trajectories,
            ) = hdbscan_clustering(
                min_cluster_size=min_cluster_size,
                min_samples=min_samples,
                clustering_matrix=self.clustering_matrix,
                frames=self.frames,
                find_centroid=True,
                trajectory_names=self.trajectory_names,
                **kwargs
            )
            sil_score = round(silhouette_score(self.clustering_matrix, labels), 4)
            print(
                f"Clustering with parameters min_cluster_size={min_cluster_size} and min_samples={min_samples} produces {len(set(labels))} clusters with an average silhouette score of {sil_score}."
            )
            silhouette_scores = pd.DataFrame(
                {
                    "Number of clusters": [len(set(labels))],
                    "Minimum cluster size": [min_cluster_size],
                    "Minimum samples": [min_samples],
                    "Average silhouette score": [sil_score],
                }
            )

        elif automatic == True:
            sil_scores = []
            params = []
            cluster_nums = []
            for min_cluster_size in np.arange(min_cluster_size_range[0],
                                              min_cluster_size_range[1],
                                              min_cluster_size_range[2]):
                for min_samples in np.arange(
                    min_samples_range[0], 
                    min_samples_range[1], 
                    min_samples_range[2]
                ):
                    labels = hdbscan_clustering(
                        min_cluster_size=min_cluster_size,
                        min_samples=min_samples,
                        clustering_matrix=self.clustering_matrix,
                        frames=self.frames,
                        find_centroid=False,
                        trajectory_names=self.trajectory_names,
                    )

                    if len(set(labels)) == len(labels):
                        print(
                            f"Clustering with parameters min_cluster_size={min_cluster_size} and min_samples={min_samples} produces no clusters. Moving on to next set of parameters"
                        )

                    elif len(set(labels)) <= 2:
                        print(
                            f"Clustering with parameters min_cluster_size={min_cluster_size} and min_samples={min_samples} produces only one cluster. Moving on to next set of parameters"
                        )

                    elif len(set(labels)) > max_clusters:
                        print(
                            f"Clustering with parameters min_cluster_size={min_cluster_size} and min_samples={min_samples} produces more than {max_clusters} clusters. Moving on to next set of parameters"
                        )

                    elif (
                        determine_cluster_population_ratio(
                            labels=labels,
                            max_cluster_population=max_cluster_population,
                            clustering_method=clustering_method,
                        )
                        == True
                    ):
                        print(
                            f"Clustering with parameters min_cluster_size={min_cluster_size} and min_samples={min_samples} results in the most populated cluster containing more than {max_cluster_population*100}% of frames. Moving on to next st of parameters."
                        )

                    else:
                        try:
                            sil_score = round(
                                silhouette_score(self.clustering_matrix, labels), 4
                            )
                            sil_scores.append(sil_score)
                            cluster_nums.append(len(set(labels)))
                            params.append([min_cluster_size, min_samples])
                            print(
                                f"Clustering with parameters min_cluster_size={min_cluster_size} and min_samples={min_samples} produces {len(set(labels))} clusters with an average silhouette score of {sil_score}."
                            )
                        except Exception as e:
                            print(f'Error found when evaluating silhouette score with min_cluster_size={min_cluster_size} and min_samples={min_samples} so skipping this: {e}.')
                            continue

            sil_scores = np.array(sil_scores)
            best_params_index = np.argmax(sil_scores)
            best_min_cluster_size, best_min_samples = tuple(params[best_params_index])
            sil_score = sil_scores[best_params_index]
            print(
                f"Best clustering parameters identified at min_cluster_size={min_cluster_size} and min_samples={best_min_samples} with an average silhouette score of {sil_score}."
            )
            (
                labels,
                cluster_centers,
                cluster_center_indices,
                cluster_centers_trajectories,
            ) = hdbscan_clustering(
                min_cluster_size=best_min_cluster_size,
                min_samples=best_min_samples,
                clustering_matrix=self.clustering_matrix,
                frames=self.frames,
                find_centroid=True,
                trajectory_names=self.trajectory_names,
                **kwargs
            )
            params = np.array(params)

            silhouette_scores = pd.DataFrame(
                {
                    "Number of clusters": cluster_nums,
                    "Minimum cluster size": params[:, 0],
                    "Minimum samples": params[:, 1],
                    "Average silhouette score": sil_scores,
                }
            )
        
        summary = generate_clustering_summary(
            trajectory_file=self.trajectory_file,
            topology_file=self.topology_file,
            pka_file=self.pka_file,
            selections=self.selections,
            include_distances=self.include_distances,
            include_buriedness=self.include_buriedness,
            clustering_method=clustering_method,
            automatic=automatic,
            silhouette_scores=silhouette_scores,
            n_components=self.n_components,
            cummulative_variance=self.cummulative_variance,
            buriedness_file=self.buriedness_file,
        )
        self._labels, self._cluster_centers, self._cluster_center_indices, self._cluster_centers_trajectories  = (
            write_clustering_info(
            summary=summary,
            trajectory_dict=self.trajectory_dict,
            pka_df=self.pka_df,
            times=self.times,
            frames=self.frames,
            trajectory_names=self.trajectory_names,
            labels=labels,
            cluster_centers=cluster_centers,
            cluster_center_indices=cluster_center_indices,
            cluster_centers_trajectories=cluster_centers_trajectories,
            output_directory=self.output_directory,
            output_prefix=self.output_prefix,
            clustering_method=clustering_method,
            )
        )


"""
    @release_date  : $release_date
    @version       : $release_version
    @author        : Christos Matsingos, Ka Fu Man 
    
    This file is part of the TrIPP software
    (https://github.com/fornililab/TrIPP).
    Copyright (c) 2024 Christos Matsingos, Ka Fu Man and Arianna Fornili.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

from tripp._sort_clusters_ import sort_clusters 
import pandas as pd 
import MDAnalysis as mda 

def write_clustering_info(summary, trajectory_dict, pka_df, times, frames, trajectory_names, labels, cluster_centers, cluster_center_indices, cluster_centers_trajectories, output_directory, output_prefix, clustering_method): 
    """
    Writes clustering information to a log file and saves cluster centers as PDB files.
    Parameters
    ----------
    summary : str
        A summary string for the clustering parameters generated by generate_clustering_summary.
    trajectory_dict : dict
        A dictionary where the keys are trajectory names and the values are MDAnalysis Universe objects.
    pka_df : pd.DataFrame
        A DataFrame containing pKa values with rows indexed by time and columns by residue identifier.
    times : np.ndarray
        An array of simulation times corresponding to the frames provided as input for clustering.
    frames : np.ndarray
        An array of input frames.
    trajectory_names : np.ndarray
        An array of trajectories names for the input frames.
    labels : np.ndarray
        An array of cluster labels for each frame.
    cluster_centers: list
        Indices of the cluster centers within their individual trajectories (local indices). 
        Indices start from 0 for each trajectory.
    cluster_center_indices: list
        Indices of the cluster centers within the full feature matrix (global indices).
    cluster_centers_trajectories: list
        Trajectory names of the cluster centers.
    output_directory : str
        The path to the directory where output files will be saved.
    output_prefix : str
        Output file prefix.
    clustering_method : str
        The clustering method used (e.g., 'DBSCAN', 'KMedoids').
    Returns
    -------
    labels: np.ndarray
        An array of cluster labels for each point in the feature matrix.
    cluster_centers: list
        Indices of the cluster centers within their individual trajectories (local indices), 
        sorted by cluster population 
        Indices start from 0 for each trajectory.
    cluster_center_indices: list
        Indices of the cluster centers within the full feature matrix (global indices),
        sorted by cluster population
    cluster_centers_trajectories: list
        Trajectory names of the cluster centers, sorted by cluster population
    """
    labels, cluster_centers, cluster_center_indices, cluster_centers_trajectories = sort_clusters(labels=labels, cluster_centers=cluster_centers, cluster_center_indices=cluster_center_indices, cluster_centers_trajectories=cluster_centers_trajectories)

    def write_clustering_log():
        df = pd.DataFrame({'Times' : times.flatten(), 'Frames' : frames.flatten(), 'Labels' : labels.flatten(), 'Trajectories' : trajectory_names.flatten()}) 
        cluster_data = {} 
        for i in range(len(cluster_centers)): 
            df_i = df[df['Labels']==i]
            cluster_data[f'Cluster {i}'] = {'Population' : f'{round((len(df_i)/len(df))*100,2)}%', 
                                            'Centroid frame' : str(cluster_centers[i]), 
                                            'Centroid time' : str(times.flatten()[cluster_center_indices[i]]), 
                                            'Centroid trajectory' : str(trajectory_names.flatten()[cluster_center_indices[i]])} 
            for traj in trajectory_dict.keys(): 
                cluster_data[f'Cluster {i}'][traj] = {'Times' : ', '.join(map(str, df_i[df_i['Trajectories']==traj]['Times'])), 
                                                      'Frames' : ', '.join(map(str, df_i[df_i['Trajectories']==traj]['Frames']))} 
        
        if clustering_method == 'DBSCAN': 
            df_i = df[df['Labels']==-1]
            cluster_data['Cluster -1 (Outliers)'] = {'Population' : f'{round((len(df_i)/len(df))*100,2)}%'} 
            for traj in trajectory_dict.keys(): 
                cluster_data['Cluster -1 (Outliers)'][traj] = {'Times' : ', '.join(map(str, df_i[df_i['Trajectories']==traj]['Times'])), 
                                                               'Frames' : ', '.join(map(str, df_i[df_i['Trajectories']==traj]['Frames']))} 
        
        with open(f'{output_directory}/{output_prefix}_{clustering_method}.log', 'w') as l: 
            l.write(summary) 
            for i in range(len(cluster_centers)): 
                l.write(f'\nCluster {i}\n\n') 
                for key in cluster_data[f'Cluster {i}'].keys(): 
                    if type(cluster_data[f'Cluster {i}'][key]) == dict: 
                        l.write(f'{key}\n\n'+'Times'+'\n'+cluster_data[f'Cluster {i}'][key]['Times']+'\n\n') 
                        l.write(f'Frames'+'\n'+cluster_data[f'Cluster {i}'][key]['Frames']+'\n\n') 
                    else: 
                        l.write(f'{key}\t\t'+cluster_data[f'Cluster {i}'][key]+'\n\n') 
            
            if clustering_method == 'DBSCAN': 
                l.write('\nCluster -1 (Outliers)\n\n') 
                for key in cluster_data[f'Cluster -1 (Outliers)'].keys(): 
                    if type(cluster_data['Cluster -1 (Outliers)'][key]) == dict: 
                        l.write(f'{key}\n\n'+'Times'+'\n'+cluster_data['Cluster -1 (Outliers)'][key]['Times']+'\n\n') 
                        l.write(f'Frames'+'\n'+cluster_data['Cluster -1 (Outliers)'][key]['Frames']+'\n\n') 
                    else: 
                        l.write(f'{key}\t\t'+cluster_data['Cluster -1 (Outliers)'][key]+'\n\n') 
    
    write_clustering_log()  
    
    def write_cluster_centers(): 
        for index, cluster_center_frame in enumerate(cluster_centers): 
            universe = trajectory_dict[cluster_centers_trajectories[index]] 
            for ts in universe.trajectory: 
                if ts.frame == cluster_center_frame: 
                    pdb_file = f'{output_directory}/{output_prefix}_{clustering_method}_C{index}.pdb' 
                    with mda.Writer(pdb_file) as w: 
                        w.write(universe) 
    
    write_cluster_centers() 

    def write_new_dataframe(): 
        pka_df['Clusters'] = labels 
        pka_df.to_csv(f'{output_directory}/{output_prefix}_{clustering_method}_cluster.csv') 
    
    write_new_dataframe() 
    return labels, cluster_centers, cluster_center_indices, cluster_centers_trajectories



"""
    @release_date  : $release_date
    @version       : $release_version
    @author        : Christos Matsingos, Ka Fu Man

    This file is part of the TrIPP software
    (https://github.com/fornililab/TrIPP).
    Copyright (c) 2024 Christos Matsingos, Ka Fu Man and Arianna Fornili.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

from sklearn.cluster import DBSCAN
from tripp._density_clustering_utilities_ import find_density_clustering_centroids

def dbscan_clustering(
    eps,
    min_samples,
    metric,
    metric_params,
    algorithm,
    leaf_size,
    p,
    n_jobs,
    clustering_matrix,
    frames,
    find_centroid,
    trajectory_names,
):
    """
    Function to run DBSCAN clustering from scikit-learn.
    Standard DBSCAN parameters can be found in the scikit-learn documentation:
    https://scikit-learn.org/stable/modules/generated/sklearn.cluster.DBSCAN.html
    
    Parameters
    ----------
    clustering_matrix : np.ndarray
       The clustering matrix (feature matrix) created by the create_clustering_matrix function.
    frames : list
        A list of frames corresponding to the points to be clustered.
        Generated by the create_clustering_matrix function.
    find_centroid : bool
        If True, the function will return the cluster centroids.
        If False, it will only return the labels.
    trajectory_names : np.ndarray
        An array of names of the trajectories to be clustered.
        Generated by the create_clustering_matrix function.
    Returns
    -------
    labels: np.ndarray
        An array of cluster labels for each point in the feature matrix.
    
    If `find_centroid=True`:
    
    cluster_centers: list
        Indices of the cluster centers within their individual trajectories (local indices). 
        Indices start from 0 for each trajectory.
    cluster_center_indices: list
        Indices of the cluster centers within the full feature matrix (global indices).
    cluster_centers_trajectories: list
        Trajectory names of the cluster centers.
    """

    dbscan_clustering = DBSCAN(
        eps=eps,
        min_samples=min_samples,
        metric=metric,
        metric_params=metric_params,
        algorithm=algorithm,
        leaf_size=leaf_size,
        p=p,
        n_jobs=n_jobs,
    )
    dbscan_clustering.fit(clustering_matrix)
    labels = dbscan_clustering.labels_

    if find_centroid is True:
        (cluster_centers, cluster_center_indices, cluster_centers_trajectories
        ) = find_density_clustering_centroids(labels, frames, trajectory_names, clustering_matrix)
        return (
            labels,
            cluster_centers,
            cluster_center_indices,
            cluster_centers_trajectories,
        )

    elif find_centroid is False:
        return labels



"""
    @release_date  : $release_date
    @version       : $release_version
    @author        : Christos Matsingos, Ka Fu Man 
    
    This file is part of the TrIPP software
    (https://github.com/fornililab/TrIPP).
    Copyright (c) 2024 Christos Matsingos, Ka Fu Man and Arianna Fornili.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

import pandas as pd 
from tripp._model_pka_values_ import model_pka_values 
import numpy as np
from scipy import stats


def calculate_difference_to_model(input_directory, output_prefix):
    """Calculate the time evolution of the difference between PROPKA predicted
    and model pKa values.

    Parameters
    ----------
    output_directory: str
        The directory where the output CSV file will be saved.
        
    output_prefix: str
        Prefix of the output CSV file.
    
    """
    df_pka = pd.read_csv(f'{input_directory}/{output_prefix}_pka.csv')

    df_dif = pd.DataFrame()
    df_dif['Time [ps]'] = df_pka['Time [ps]']
    del df_pka['Time [ps]']
    residues = np.array(df_pka.columns)
    for residue in residues:
        if 'N+' in residue:
            residue_modified = residue.replace('N+', 'NTR')
        elif 'C-' in residue:
            residue_modified = residue.replace('C-', 'CTR')
        else:
            residue_modified = residue

        amino_acid = residue_modified[:3]
        model_value = model_pka_values[amino_acid]
        residue_data = np.around(df_pka[residue].to_numpy() - model_value, decimals=3)
        df_dif[residue] = residue_data

    df_dif.to_csv(f'{input_directory}/{output_prefix}_difference_to_model.csv', index=False)


def PCProjectionScreening(output_directory,
                          output_prefix,
                          pka_file,
                          projection_file,
                          method='Pearson',
                          start=0,
                          end=-1,
                          header=None):
    """Calculates the correlation coefficient between the property in projection_file
    and the pKa time evolution for all the ionisable residues.
    
    Parameters
    ----------
    output_directory : str
        The path of the directory where output files will be saved.
    output_prefix : str
        Output file prefix
    pka_file: str
        The path of the CSV file with the pKa values generated by the 
        Trajectory class.
    projection_file: str
        The path of the CSV file with the PC projection (or other property) values. 
        The file needs to contain the simulation time (first column) 
        and the projection value (second column) for the same frames 
        (in the same order) as in the trajectory used for the calculation 
        of the pKa values. See tutorial/files/1AKI_md1_pc1.csv for an example.
    method: str, default='Pearson'
        'Pearson' or 'Spearman' correlation.
    start: int, default=0
        Start time (ps). The correlation will be calculated using only the 
        frames with simulation time between start and end. 
        Default is 0 (beginning of the trajectory)
    end: int, default=-1
        End time (ps). The correlation will be calculated using only the frames 
        with simulation time between start and end. Default is -1 (end of the trajectory)
    header: default=None
        Default is None (no header in projection_file). 
        Use header=0 if projection_file has a header.
    """
    pKaDF = pd.read_csv(pka_file)
    ProjDF = pd.read_csv(projection_file,
                         sep=',',
                         skipinitialspace=True,
                         header=header)
    # Made sure the column name is Time [ps] and Projection.
    ProjDF.rename(columns=dict(zip(ProjDF.columns.to_list(),['Time [ps]','Projection'])), inplace=True)
    if end == -1:
        end = pKaDF['Time [ps]'].iloc[end]
    else:
        end = end
    # Slice through pKaDF and ProjDF according to the start and end provided.
    pKaDF = pKaDF[(pKaDF['Time [ps]'] >= start) & (pKaDF['Time [ps]'] <= end)]
    ProjDF = ProjDF[(ProjDF['Time [ps]'] >= start) & (ProjDF['Time [ps]'] <= end)]
    
    Correlations = []
    PValues = []
    for Residue in pKaDF.keys()[1:]:
        if method == 'Pearson':
            Correlation, PValue = stats.pearsonr(ProjDF['Projection'], pKaDF[Residue])
        elif method == 'Spearman':
            Correlation, PValue = stats.spearmanr(ProjDF['Projection'], pKaDF[Residue])
        Correlations.append(Correlation)
        PValues.append(PValue)
    OutputDF = pd.DataFrame([pKaDF.keys()[1:], Correlations, PValues]).T
    OutputDF.columns = ['Residue', 'Correlation', 'p-value']
    OutputDF.to_csv(f'{output_directory}/{output_prefix}.csv', index=False)



import numpy as np 

def determine_cluster_population_ratio(labels, max_cluster_population, clustering_method): 
    """
    Calculates the relative population (fraction of frames) for each cluster and 
    compares the maximum value to a given threshold
    Parameters
    ----------
    labels: np.ndarray
        Array where each element is the cluster ID assigned to a data point
    max_cluster_population: float
        The maximum relative population.
    clustering_method: str
        The clustering method used, e.g., 'DBSCAN'.
    Returns
    -------
    bool:
        True if the maximum relative population of any cluster is greater than 
        or equal to `max_cluster_population` (default 95%) , False otherwise.
    """
    unique_labels = list(set(labels)) 
    cluster_population = np.zeros(len(unique_labels)) 

    labels = np.ravel(labels) 

    if clustering_method == 'DBSCAN' or clustering_method == 'HDBSCAN': 
        labels = labels[labels!=-1] 

    for label in unique_labels: 
        cluster_labels = labels[labels==label] 
        cluster_elements = len(cluster_labels) 
        cluster_population[label] = cluster_elements/len(labels) 
    
    if np.max(cluster_population) >= max_cluster_population: 
        return True 
    
    else: 
        return False 


"""
    @release_date  : $release_date
    @version       : $release_version
    @author        : Christos Matsingos, Ka Fu Man 
    
    This file is part of the TrIPP software
    (https://github.com/fornililab/TrIPP).
    Copyright (c) 2024 Christos Matsingos, Ka Fu Man and Arianna Fornili.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
from tripp._calculate_charge_center_distances_ import calculate_charge_center_distances
from tripp._determine_charge_center_ import determine_charge_center
from scipy.stats import zscore
from tripp._create_mda_universe_ import create_mda_universe


def create_clustering_matrix(trajectory_file, topology_file, pka_df,
                             selections, include_distances,
                             buriedness_df, include_buriedness):
    """
    Function that generates clustering_matrix (feature matrix), where features
    include Z-score standardised pKa values (from pka_df) of selected 
    residues, together with (optionally) Z-score standardised buried ratios 
    (from buriedness_df) and inter-residue distances (calculated using the  
    calculate_charge_center_distances function). 
    
    Parameters
    ----------
    trajectory_file: str or dict
        When str, it is the path of the file containing the trajectory. The same
        formats allowed by MDAnalysis can be used. When dict, the clustering is performed using
        multiple trajectories specified in a dictionary, where trajectory names are
        used as key and file paths as values, e.g.: {'MD1' : 'file1 path', 'MD2' : 'file2 path', ...}.
        The same topology file is used for all trajectories.
    topology_file: str
        Path to the topology file. The same formats allowed by MDAnalysis can be used.
    pka_df: pd.DataFrame
        A DataFrame containing pKa values labelled by time, residue identifier, and
        trajectory name. This DataFrame is processed by the make_pka_or_buriendess_df function
        in the Clustering class.
    selections: list of str
        List of residues used for the clustering (selected with MDAnalysis selection syntax)
        For multi-chain systems, your topology and selection must include chain IDs.
    include_distances: bool
        If True, inter-residue distances (calculated between the charge centers) are used as
        additional features for the clustering alongside the pKa values.
    buriedness_df: pd.DataFrame
        A DataFrame containing buried ratio (buriedness) values labelled by time, residue identifier,
        and trajectory name. This DataFrame is processed by the make_pka_or_buriendess_df function
        in the Clustering class.
    include_buriedness: bool
        If True, buried ratio values will be included in the feature matrix.
        Default is False.
    Returns
    -------
    clustering_matrix: np.ndarray
        A 2D numpy array containing the normalized pKa values and distances.
    times: np.ndarray
        A 1D numpy array containing the times corresponding to the clustering data.
    frames: np.ndarray
        A 1D numpy array containing the frames corresponding to the clustering data.
    trajectory_names: np.ndarray
        A 1D numpy array containing the names of the trajectories corresponding to the clustering data.
    trajectory_dict: dict
        A dictionary where keys are trajectory names and values are MDAnalysis Universe objects.
    """

    if isinstance(trajectory_file, str):
        trajectory_dict = {
            "Unnamed trajectory": create_mda_universe(
                topology_file=topology_file,
                trajectory_file=trajectory_file
                )
            }

    elif isinstance(trajectory_file, dict):
        trajectory_dict = {}
        for key in trajectory_file.keys():
            traj = trajectory_file[key]
            trajectory_dict[key] = create_mda_universe(
                topology_file=topology_file, trajectory_file=traj
            )

    for index, key in enumerate(trajectory_dict.keys()):
        df_traj_pka = pka_df[pka_df["Trajectories"] == key]
        if include_buriedness:
            df_traj_buriedness = buriedness_df[buriedness_df["Trajectories"] == key]
        elif not include_buriedness:
            df_traj_buriedness = None
        if index == 0:
            clustering_matrix, times, frames, trajectory_names = (
                extract_pka_distances_buriedness(key,
                                                 trajectory_dict[key],
                                                 df_traj_pka,
                                                 df_traj_buriedness,
                                                 selections, 
                                                 include_distances,
                                                 include_buriedness)
            )

        else:
            partial_clustering_matrix, partial_times, partial_frames, partial_trajectory_names = (
                extract_pka_distances_buriedness(key, 
                                                 trajectory_dict[key],
                                                 df_traj_pka,
                                                 df_traj_buriedness,
                                                 selections,
                                                 include_distances,
                                                 include_buriedness))
            clustering_matrix = np.concatenate((clustering_matrix, partial_clustering_matrix), axis=0)
            times = np.concatenate((times, partial_times), axis=0)
            frames = np.concatenate((frames, partial_frames), axis=0)
            trajectory_names = np.concatenate((trajectory_names, partial_trajectory_names), axis=0)

    return clustering_matrix, times, frames, trajectory_names, trajectory_dict


def extract_pka_distances_buriedness(trajectory_name, universe,
                                     df_traj_pka, df_traj_buriedness,
                                     selections, include_distances,
                                     include_buriedness):

        pka = []
        times = []
        frames = []
        trajectory_names = []
        d = []
        b = []

        for ts in universe.trajectory:
            times.append([ts.time])
            frames.append([ts.frame])
            trajectory_names.append(trajectory_name)
            positions = []
            pkas = []
            buriedness = []
            for selection in selections:
                charge_center, residue_identifier = determine_charge_center(
                                    universe, selection
                                )
                
                pka_residue = df_traj_pka.loc[ts.time, residue_identifier]
                pkas.append(pka_residue)

                if include_distances is True:
                    positions.append(charge_center)

                if include_buriedness is True:
                    buriedness_residue = df_traj_buriedness.loc[
                        ts.time, residue_identifier
                    ]
                    buriedness.append(buriedness_residue)

            pka.append(np.array(pkas))

            if include_distances is True:
                distances = calculate_charge_center_distances(positions)
                d.append(distances)

            if include_buriedness is True:
                b.append(buriedness)

        pka = np.array(pka)

        if include_distances is True and include_buriedness is True:
            d = np.array(d)
            b = np.array(b)
            clustering_matrix = np.concatenate(
                (zscore(d, axis=None), zscore(b, axis=None), zscore(pka, axis=None)),
                axis=1,
            )

        elif include_distances is True and include_buriedness is False:
            d = np.array(d)
            clustering_matrix = np.concatenate(
                (zscore(d, axis=None), zscore(pka, axis=None)), axis=1
            )

        elif include_distances is False and include_buriedness is True:
            b = np.array(b)
            clustering_matrix = np.concatenate(
                (zscore(b, axis=None), zscore(pka, axis=None)), axis=1
            )

        elif include_distances is False and include_buriedness is False:
            clustering_matrix = zscore(pka, axis=None)

        times = np.array(times)
        frames = np.array(frames)
        trajectory_names = np.array(trajectory_names)

        return clustering_matrix, times, frames, trajectory_names



"""
    @release_date  : $release_date
    @version       : $release_version
    @author        : Christos Matsingos, Ka Fu Man 
    
    This file is part of the TrIPP software
    (https://github.com/fornililab/TrIPP).
    Copyright (c) 2024 Christos Matsingos, Ka Fu Man and Arianna Fornili.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

import MDAnalysis as mda 
import numpy as np 

def determine_charge_center(universe, selection):
    """
    Determines the charge center of a residue in a MDAnalysis universe.
    Parameters
    ----------
    universe: MDAnalysis.universe
    
    selection: str
        A string (MDAnalysis selection syntax) to select the residue for
        charge center calculation.
    Returns
    -------
    charge_center: np.ndarray
        The charge center coordinates for the selected residue.
    residue_identifier: str
        A string identifier for the residue, formatted as 'RESID:CHAINID'.
        The following residue types are recognized:
        - ARG, ASP, CYS, GLU, HIS, LYS, TYR
    Raises
    ------
    Exception
        If the residue type is not recognized by TrIPP, an exception is raised.
    """
    ag = universe.select_atoms(selection)
    residue_type = ag.residues.resnames
    resid = ag.residues.resids[0]
    resindex = ag.residues.resindices[0]
    chain = ag.atoms.chainIDs[0]

    # Charge center is determined as in PROPKA3
    if residue_type in ['ARG', 'ARGN', 'CARG', 'NARG']:
        atom_coordinates = universe.select_atoms(f'resindex {resindex} and name CZ').positions
        charge_center = atom_coordinates[0]
        residue_identifier = f'ARG{resid}:{chain}'

    elif residue_type in ['ASP', 'ASPH', 'ASPP', 'CASF', 'CASP', 'NASP', 'ASF', 'ASH']:
        atom_coordinates = universe.select_atoms(f'resindex {resindex} and name OD1 OD2').positions
        charge_center = np.mean(atom_coordinates, axis=0)
        residue_identifier = f'ASP{resid}:{chain}'

    elif residue_type in ['CYS', 'CCYS', 'CCYX', 'CYS1', 'CYS2', 'CYSH', 'NCYS', 'NCYX', 'CYM', 'CYN', 'CYX']:
        atom_coordinates = universe.select_atoms(f'resindex {resindex} and name SG').positions
        charge_center = atom_coordinates[0]
        residue_identifier = f'CYS{resid}:{chain}'

    elif residue_type in ['GLU', 'CGLU', 'GLUH', 'GLUP', 'NGLU', 'PGLU', 'GLH']:
        atom_coordinates = universe.select_atoms(f'resindex {resindex} and name OE1 OE2').positions
        charge_center = np.mean(atom_coordinates, axis=0) 
        residue_identifier = f'GLU{resid}:{chain}'

    elif residue_type in ['HIS', 'CHID', 'CHIE', 'CHIP', 'HIS1', 'HIS2', 'HISA', 'HISB', 'HISD', 'HISE', 'HISH', 'NHID', 'NHIE', 'NHIP', 'HID', 'HIE', 'HIP', 'HSD', 'HSE', 'HSP']:
        atom_coordinates = universe.select_atoms(f'resindex {resindex} and name CG CD2 ND1 CE1 NE2').positions
        charge_center = np.mean(atom_coordinates, axis=0)
        residue_identifier = f'HIS{resid}:{chain}'

    elif residue_type in ['LYS', 'CLYS', 'LYSH', 'NLYS', 'LYN', 'LSN']:
        atom_coordinates = universe.select_atoms(f'resindex {resindex} and name NZ').positions
        charge_center = atom_coordinates[0]
        residue_identifier = f'LYS{resid}:{chain}'

    elif residue_type in ['TYR', 'CTYR', 'NTYR']:
        atom_coordinates = universe.select_atoms(f'resindex {resindex} and name OH').positions
        charge_center = atom_coordinates[0]
        residue_identifier = f'TYR{resid}:{chain}'

    else:
        raise Exception(f'Residue {residue_type}{resid}:{chain} is not recognized by TrIPP: unable to determine charge centre')

    if np.any(np.isnan(charge_center)):
        raise Exception(f'Unable to determine charge center for {residue_type}{resid}:{chain} due to missing atoms in the residue.')

    return charge_center, residue_identifier


"""
    @release_date  : $release_date
    @version       : $release_version
    @author        : Christos Matsingos, Ka Fu Man

    This file is part of the TrIPP software
    (https://github.com/fornililab/TrIPP).
    Copyright (c) 2024 Christos Matsingos, Ka Fu Man and Arianna Fornili.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
from tqdm import tqdm

def calculate_rmsd_matrix(clustering_matrix, frames):
    """
    Function that calculates the pairwise Euclidean distance matrix for the greedy clustering algorithm.
    
    Parameters
    ----------
    clustering_matrix : np.ndarray
        The clustering matrix (feature matrix) created by the create_clustering_matrix function.
    frames : list
        A list of frames corresponding to the points to be clustered.
    Returns
    -------
    rmsd_array : np.ndarray
        A 2D numpy array containing the Euclidean distance values between each pair of points.
    """

    def calculate_rmsd(point1, point2):

        sqr_sum = np.sum(np.square(point1 - point2))
        rmsd = np.sqrt(sqr_sum)

        return rmsd

    rmsd_array = np.zeros((len(frames), len(frames)))
    print('Building distance matrix...')
    for i in tqdm(range(len(frames))):
        point1 = clustering_matrix[i]
        for j in range(len(frames)):
            point2 = clustering_matrix[j]
            rmsd_array[i, j] = calculate_rmsd(point1, point2)

    return rmsd_array



"""
    @release_date  : $release_date
    @version       : $release_version
    @author        : Christos Matsingos, Ka Fu Man

    This file is part of the TrIPP software
    (https://github.com/fornililab/TrIPP).
    Copyright (c) 2024 Christos Matsingos, Ka Fu Man and Arianna Fornili.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np 

def find_smallest_distance(cluster, frames, labels, trajectory_names, clustering_matrix):
    """Find the point in the cluster that is closest to the centroid of the cluster.
    Parameters
    ----------
    cluster : list
        A list of unique cluster labels.
    frames : list
        A list of frames corresponding to feature matrix points.
    labels : np.ndarray
        An array of cluster labels for each point in the feature matrix.
    trajectory_names : np.ndarray
        A 1D array containing the name of each frame in the feature matrix.
    clustering_matrix : np.ndarray
        The clustering matrix (feature matrix) created by the create_clustering_matrix function.
    Returns
    -------
    center : int
        The index of the centroid frame in its individual trajectory.
    center_index : int
        The index of the centroid frame in the full feature matrix (i.e., global index).
    center_trajectory : str
        The name of the trajectory to which the centroid frame belongs.
    """
    # The centroid is defined as the point which has the smallest
    # distance from the mean of all points in the cluster.
    cluster_frames = frames[labels == cluster]
    cluster_trajectories = trajectory_names[labels == cluster]
    cluster_members = clustering_matrix[labels == cluster]
    cluster_mean = np.mean(cluster_members, axis=0)
    dist_mean = []
    for item in cluster_members:
        dist_mean.append(np.sqrt(np.sum(np.square(item - cluster_mean))))
    smallest_index = np.argmin(np.array(dist_mean))
    center = cluster_frames[smallest_index][0]
    center_trajectory = cluster_trajectories[smallest_index]
    center_index_frame_mask = np.ravel(frames == center)
    center_index_trajectory_mask = trajectory_names == center_trajectory
    center_index_mask = center_index_frame_mask & center_index_trajectory_mask
    center_index = np.where(center_index_mask)[0][0]
    return center, center_index, center_trajectory

def find_density_clustering_centroids(labels, frames, trajectory_names, clustering_matrix):
    """Find the centroids of the density clusters.
    Parameters
    ----------
    labels : np.ndarray
        An array of cluster labels for each point in the feature matrix.
    frames : list
        A list of frames corresponding to feature matrix points.
    trajectory_names : np.ndarray
        A 1D array containing the name of each frame in the feature matrix.
    clustering_matrix : np.ndarray
        The clustering matrix (feature matrix) created by the create_clustering_matrix function.
    Returns
    -------
    cluster_centers: list
        The indices of the cluster centers within their respective individual trajectories.
    cluster_center_indices: list
        The indices of the cluster centers in the full feature matrix (i.e., global indices).
    cluster_centers_trajectories: list
         A list of trajectory names of the cluster centers.
    """
    
    # A list of clusters is made to iterate over.
    clusters = list(set(labels))

    if -1 in clusters:
        # The cluster label -1 is removed since it is the label
        # for outliers.
        clusters.remove(-1)

    cluster_centers = []
    cluster_center_indices = []
    cluster_centers_trajectories = []

    for cluster in clusters:
        center, center_index, center_trajectory = find_smallest_distance(cluster, frames, labels, trajectory_names, clustering_matrix)
        cluster_centers.append(center)
        cluster_center_indices.append(center_index)
        cluster_centers_trajectories.append(center_trajectory)

    return cluster_centers, cluster_center_indices, cluster_centers_trajectories